# Check the following urls for more info about Pan-STARRS:
#
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service#PS1ImageCutoutService-ImportantFITSimageformat,WCS,andflux-scalingnotes
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images#PS1Stackimages-Photometriccalibration
#
# For DES:
#
#     https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/processing
#
# For SDSS:
#
#     https://www.sdss.org/dr12/algorithms/fluxcal/#SDSStoAB
#     https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
#
# Some parts of this notebook are based on https://github.com/djones1040/PS1_surface_brightness/blob/master/Surface%20Brightness%20Tutorial.ipynb and codes from Lluís Galbany

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import sep
from photutils.aperture import aperture_photometry, CircularAperture

from astropy.io import fits
from astropy import units as u, wcs
from astropy.cosmology import FlatLambdaCDM

from hostphot._constants import workdir
from hostphot.utils import (
    get_survey_filters,
    check_survey_validity,
    check_filters_validity,
    calc_sky_unc,
    get_image_exptime,
    survey_pixel_scale,
    check_work_dir,
    update_axislabels,
    magnitude_calculation,
    bkg_surveys,
)
from hostphot.image_cleaning import remove_nan
from hostphot.dust import calc_extinction

import warnings
from astropy.utils.exceptions import AstropyWarning

H0 = 70
Om0 = 0.3
__cosmo__ = FlatLambdaCDM(H0, Om0)
sep.set_sub_object_limit(1e4)

# ----------------------------------------
def choose_cosmology(cosmo):
    """Updates the cosmology used to calculate the aperture size.

    Parameters
    ----------
    cosmo: `astropy.cosmology` object
        Cosmological model. E.g. :func:`FlatLambdaCDM(70, 0.3)`.
    """
    global __cosmo__
    __cosmo__ = cosmo


# -------------------------------
def calc_aperture_size(z, ap_radius):
    """Calculates the size of the aperture in arsec,
    for aperture photometry, given a physical size.

    Parameters
    ----------
    z: float
        Redshift.
    ap_radius: float
        Physical aperture size in kpc.

    Returns
    -------
    radius_arcsec: float
        Aperture size in arcsec.
    """
    ap_radius = ap_radius * u.kpc

    # transverse separations
    transv_sep_per_arcmin = __cosmo__.kpc_proper_per_arcmin(z)
    transv_sep_per_arcsec = transv_sep_per_arcmin.to(u.kpc / u.arcsec)

    radius_arcsec = ap_radius / transv_sep_per_arcsec

    return radius_arcsec.value


def extract_aperture_flux(data, error, px, py, radius):
    """Extracts aperture photometry of a single image.

    Parameters
    ----------
    data: array
        Image data in a 2D numpy array.
    error: array
        Errors of ``data``.
    px: float
        x-axis pixel coordinate of the aperture center.
    py: float
        y-axis pixel coordinate of the aperture center.
    radius: float
        Aperture radius in pixels.

    Returns
    -------
    raw_flux: float
        Aperture photometry ("raw" flux).
    raw_flux_err: float
        Uncertainty on the aperture photometry.
    """
    aperture = CircularAperture((px, py), r=radius)
    ap_results = aperture_photometry(data, aperture, error=error)
    raw_flux = ap_results["aperture_sum"][0]
    raw_flux_err = ap_results["aperture_sum_err"][0]

    return raw_flux, raw_flux_err


def plot_aperture(data, px, py, radius_pix, img_wcs, outfile=None):
    """Plots the aperture for the given parameters.

    Parameters
    ----------
    data: ndarray
        Data of an image.
    px: float
        X-axis center of the aperture in pixels.
    py: float
        Y-axis center of the aperture in pixels.
    radius_pix: float
        Aperture radius in pixels.
    img_wcs: WCS
        Image's WCS.
    outfile: str, default ``None``
        If given, path where to save the output figure.
    """
    m, s = np.nanmean(data), np.nanstd(data)

    fig = plt.figure(figsize=(10, 10))
    ax = plt.subplot(projection=img_wcs)
    update_axislabels(ax)
    ax.scatter(px, py, marker="*", s=140, c="g", edgecolor="gold")

    im = ax.imshow(
        data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )

    circle = plt.Circle(
        (px, py), radius_pix, color="r", fill=False, linewidth=1.5
    )
    ax.add_patch(circle)

    if outfile:
        basename = os.path.basename(outfile)
        title = os.path.splitext(basename)[0]
        title = "-".join(part for part in title.split("_"))
        fig.suptitle(title, fontsize=28)
        plt.tight_layout()
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.show()


def photometry(
    name,
    ra,
    dec,
    z,
    filt,
    survey,
    ap_radii=1,
    bkg_sub=None,
    use_mask=True,
    correct_extinction=True,
    save_plots=True,
):
    """Calculates the local aperture photometry in a given radius.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    ra: float
        Right Ascensions in degrees to center the aperture.
    dec: float
        Declinations in degrees to center the aperture.
    z: float
        Redshift of the object to estimate the physical calculate
        of the aperture.
    filt: str
        Filter to use to load the fits file.
    survey: str
        Survey to use for the zero-points and pixel scale.
    ap_radii: float or list-like, default ``1``
        Physical size of the aperture in kpc.
    bkg_sub: bool, default ``None``
        If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    use_mask: bool, default ``True``
        If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: bool, default `True`
        If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    save_plots: bool, default ``True``
        If ``True``, the figure with the aperture is saved.

    Returns
    -------
    mags: list
        Aperture magnitudes for the given aperture radii.
    mags_err: list
        Aperture magnitude errors for the given aperture radii.
    fluxes: list
        Aperture flux for the given aperture radii.
    fluxes_err: list
        Aperture flux errors for the given aperture radii.
    """
    check_survey_validity(survey)
    check_work_dir(workdir)
    obj_dir = os.path.join(workdir, name)
    if use_mask:
        suffix = "masked_"
    else:
        suffix = ""
    fits_file = os.path.join(obj_dir, f"{suffix}{survey}_{filt}.fits")

    img = fits.open(fits_file)
    img = remove_nan(img)

    header = img[0].header
    data = img[0].data
    exptime = get_image_exptime(header, survey)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    # turn float into a list
    if isinstance(ap_radii, (float, int)):
        ap_radii = [ap_radii]

    mags, mags_err = [], []
    fluxes, fluxes_err = [], []
    px, py = img_wcs.wcs_world2pix(ra, dec, 1)
    pixel_scale = survey_pixel_scale(survey, filt)
    error = calc_sky_unc(data_sub, exptime)

    for ap_radius in ap_radii:
        # aperture photometry
        radius_arcsec = calc_aperture_size(z, ap_radius)
        radius_pix = radius_arcsec / pixel_scale

        flux, flux_err = extract_aperture_flux(
            data_sub, error, px, py, radius_pix
        )

        ap_area = 2 * np.pi * (radius_pix**2)
        mag, mag_err, flux, flux_err = magnitude_calculation(
            flux,
            flux_err,
            survey,
            filt,
            ap_area,
            header,
            bkg_rms,
        )

        # extinction correction is optional
        if correct_extinction:
            A_ext = calc_extinction(filt, survey, ra, dec)
            mag -= A_ext

        mags.append(mag)
        mags_err.append(mag_err)
        fluxes.append(flux)
        fluxes_err.append(flux_err)

        if save_plots:
            outfile = os.path.join(
                obj_dir, f"local_{survey}_{filt}_{ap_radius}kpc.jpg"
            )
            plot_aperture(data_sub, px, py, radius_pix, img_wcs, outfile)

    return mags, mags_err, fluxes, fluxes_err


def multi_band_phot(
    name,
    ra,
    dec,
    z,
    filters=None,
    survey="PS1",
    ap_radii=1,
    bkg_sub=None,
    use_mask=True,
    correct_extinction=True,
    save_plots=True,
    save_results=True,
    raise_exception=False,
):
    """Calculates the local aperture photometry for multiple filters.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    ra: float
        Right Ascensions in degrees to center the aperture.
    dec: float
        Declinations in degrees to center the aperture.
    z: float
        Redshift of the object to estimate the physical calculate
        of the aperture.
    filters: str, default, ``None``
        Filters to use to load the fits files. If ``None`` use all
        the filters of the given survey.
    survey: str, default ``PS1``
        Survey to use for the zero-points and pixel scale.
    ap_radii: float or list-like, default ``1``
        Physical size of the aperture in kpc.
    bkg_sub: bool, default ``None``
        If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    use_mask: bool, default ``True``
        If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: bool, default `True`
        If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    save_plots: bool, default ``True``
        If ``True``, the figure with the aperture is saved.
    save_results: bool, default ``True``
        If ``True``, the magnitudes are saved into a csv file.
    raise_exception: bool, default ``False``
        If ``True``, an exception is raised if the photometry fails for any filter.

    Returns
    -------
    results_dict: dict
        Dictionary with the object's photometry and other info.

    Examples
    --------
    >>> import hostphot.local_photometry as lp
    >>> ap_radii = [3, 4]  # aperture radii in units of kpc
    >>> ra, dec =  308.22579, 9.92853 # coords of SN2004eo
    >>> z = 0.0157  # redshift
    >>> survey = 'PS1'
    >>> results = lp.multi_band_phot(name, ra, dec, z,
                            survey=survey, ap_radii=ap_radii,
                            use_mask=True, save_plots=True)
    """
    check_survey_validity(survey)
    if filters is None:
        if survey=='HST':
            raise ValueError("For HST, the filter needs to be specified!")
        filters = get_survey_filters(survey)
    else:
        check_filters_validity(filters, survey)
    if survey=='HST':
        filters = [filters]

    # turn float into a list
    if isinstance(ap_radii, (float, int)):
        ap_radii = [ap_radii]

    results_dict = {
        "name": name,
        "ra": ra,
        "dec": dec,
        "redshift": z,
        "survey": survey,
    }
    
    for filt in filters:
        try:
            mags, mags_err, fluxes, fluxes_err = photometry(
                name,
                ra,
                dec,
                z,
                filt,
                survey,
                ap_radii,
                bkg_sub,
                use_mask,
                correct_extinction,
                save_plots,
            )
            for radius, mag, mag_err, flux, flux_err in zip(
                ap_radii, mags, mags_err, fluxes, fluxes_err
            ):
                results_dict[f"{filt}_{radius}"] = mag
                results_dict[f"{filt}_{radius}_err"] = mag_err
                results_dict[f"{filt}_{radius}_flux"] = flux
                results_dict[f"{filt}_{radius}_flux_err"] = flux_err
        except Exception as exc:
            if raise_exception is True:
                raise Exception(exc)
            else:
                for radius in ap_radii:
                    results_dict[f"{filt}_{radius}"] = np.nan
                    results_dict[f"{filt}_{radius}_err"] = np.nan
                    results_dict[f"{filt}_{radius}_flux"] = np.nan
                    results_dict[f"{filt}_{radius}_flux_err"] = np.nan

    if save_results is True:
        outfile = os.path.join(workdir, name, f"{survey}_local.csv")
        phot_df = pd.DataFrame(
            {key: [val] for key, val in results_dict.items()}
        )
        phot_df.to_csv(outfile, index=False)

    return results_dict



import unittest
from hostphot.cutouts import download_images
from hostphot.interactive_aperture import InteractiveAperture


class TestHostPhot(unittest.TestCase):
    def test_interactivity(self):
        sn_name = "2002fk"
        ra = 50.527333
        dec = -15.400056
        survey = "PS1"

        download_images(sn_name, ra, dec, survey=survey)
        obj = InteractiveAperture(sn_name, masked=False)


if __name__ == "__main__":
    unittest.main()



# Check the following urls for more info about Pan-STARRS:
#
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service#PS1ImageCutoutService-ImportantFITSimageformat,WCS,andflux-scalingnotes
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images#PS1Stackimages-Photometriccalibration
#
# For DES:
#
#     https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/processing
#
# For SDSS:
#
#     https://www.sdss.org/dr12/algorithms/fluxcal/#SDSStoAB
#     https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
#
# Some parts of this notebook are based on https://github.com/djones1040/PS1_surface_brightness/blob/master/Surface%20Brightness%20Tutorial.ipynb and codes from Lluís Galbany

import os
import pickle
import numpy as np
import pandas as pd

import sep
from astropy import wcs
from astropy.io import fits

from hostphot._constants import workdir
from hostphot.utils import (
    get_survey_filters,
    check_survey_validity,
    check_filters_validity,
    get_image_gain,
    check_work_dir,
    magnitude_calculation,
    adapt_aperture,
    bkg_surveys,
)
from hostphot.objects_detect import extract_objects, plot_detected_objects
from hostphot.image_cleaning import remove_nan
from hostphot.dust import calc_extinction

import warnings
from astropy.utils.exceptions import AstropyWarning

sep.set_sub_object_limit(1e4)


# -------------------------------
def kron_flux(data, err, gain, objects, kronrad, scale):
    """Calculates the Kron flux.

    Parameters
    ----------
    data: ndarray
        Data of an image.
    err: float or 2D array
        Background error of the images.
    gain: float
        Gain value.
    objects: array
        Objects detected with `sep.extract()`.
    kronrad: float
        Kron radius.
    scale: float
        Scale of the Kron radius.

    Returns
    -------
    flux: array
        Kron flux.
    flux_err: array
        Kron flux error.
    """
    r_min = 1.75  # minimum diameter = 3.5

    if kronrad * np.sqrt(objects["a"] * objects["b"]) < r_min:
        print(f"Warning: using circular photometry")
        flux, flux_err, _ = sep.sum_circle(
            data,
            objects["x"],
            objects["y"],
            r_min,
            err=err,
            subpix=5,
            gain=1.0,
        )
    else:
        # theta must be in the range [-pi/2, pi/2] for sep.sum_ellipse()
        if objects["theta"] > np.pi / 2:
            objects["theta"] -= np.pi
        elif objects["theta"] < -np.pi / 2:
            objects["theta"] += np.pi

        flux, flux_err, _ = sep.sum_ellipse(
            data,
            objects["x"],
            objects["y"],
            objects["a"],
            objects["b"],
            objects["theta"],
            scale * kronrad,
            err=err,
            subpix=5,
            gain=gain,
        )

    return flux, flux_err


def optimize_kron_flux(data, err, gain, objects, eps=0.0001):
    """Optimizes the Kron flux by iteration over different values.
    The stop condition is met when the change in flux is less that ``eps``.

    Parameters
    ----------
    data: ndarray
        Data of an image.
    err: float or ndarray
        Background error of the images.
    gain: float
        Gain value.
    objects: array
        Objects detected with :func:`sep.extract()`.
    eps: float, default ``0.0001`` (0.1%)
        Minimum percent change in flux allowed between iterations.

    Returns
    -------
    opt_flux: float
        Optimized Kron flux.
    opt_flux_err: float
        Optimized Kron flux error.
    opt_kronrad: float
        Optimized Kron radius.
    opt_scale: float
        Optimized scale of the Kron radius.
    """
    # iterate over kron radii
    for r in np.arange(1, 6.05, 0.05)[::-1]:
        kronrad, _ = sep.kron_radius(
            data,
            objects["x"],
            objects["y"],
            objects["a"],
            objects["b"],
            objects["theta"],
            r,
        )
        opt_kronrad = kronrad
        if ~np.isnan(opt_kronrad):
            break

    if np.isnan(opt_kronrad):
        raise ValueError(
            "The Kron radius cannot be calculated. The image might have NaNs or the aperture is too close to the edge."
        )

    opt_flux = 0.0
    # iterate over scale
    scales = np.arange(1, 10, 0.01)
    for scale in scales:
        flux, flux_err = kron_flux(
            data, err, gain, objects, opt_kronrad, scale
        )
        flux, flux_err = flux[0], flux_err[0]

        calc_eps = np.abs(opt_flux - flux) / flux
        opt_flux, opt_flux_err = flux, flux_err
        opt_scale = scale
        if calc_eps < eps:
            opt_scale = scale
            break
        elif np.isnan(calc_eps):
            opt_scale = scales[-2]
            warnings.warn("Warning: the aperture might not fit in the image!")
            break

    return opt_flux, opt_flux_err, opt_kronrad[0], opt_scale


def extract_kronparams(
    name,
    host_ra,
    host_dec,
    filt,
    survey,
    ra=None,
    dec=None,
    bkg_sub=False,
    threshold=10,
    use_mask=True,
    optimize_kronrad=True,
    eps=0.0001,
    gal_dist_thresh=-1,
    save_plots=True,
    save_aperture_params=True,
):
    """Calculates the aperture parameters for common aperture.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy declination of the galaxy in degrees.
    filt: str or list
        Filter to use to load the fits file. List is commonly used for coadds.
    survey: str
        Survey to use for the zero-points and pixel scale.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: bool, default `None`
        If `True`, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: float, default `10`
        Threshold used by `sep.extract()` to extract objects.
    use_mask: bool, default `True`
        If `True`, the masked fits files are used. These must have
        been created beforehand.
    optimize_kronrad: bool, default `True`
        If `True`, the Kron radius is optimized, increasing the
        aperture size until the flux does not increase.
    eps: float, default ``0.0001``
        The Kron radius is increased until the change in flux is lower than ``eps``.
        A value of 0.0001 means 0.01% change in flux.
    gal_dist_thresh: float, default ``-1``.
        Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    save_plots: bool, default `True`
        If `True`, the mask and galaxy aperture figures are saved.
    save_aperture_params: bool, default `True`
        If `True`, the extracted mask parameters are saved into a pickle file.

    Returns
    -------
    gal_obj: array
        Galaxy object.
    img_wcs: WCS
        Image's WCS.
    kronrad: float
        Kron radius.
    scale: float
        Scale for the Kron radius.
    flip: bool
        Whether to flip the orientation of the
        aperture. Only used for DES images.
    """
    check_survey_validity(survey)
    check_work_dir(workdir)
    obj_dir = os.path.join(workdir, name)
    if use_mask:
        suffix = "masked_"
    else:
        suffix = ""
    if isinstance(filt, list):
        filt = "".join(f for f in filt)
    fits_file = os.path.join(obj_dir, f"{suffix}{survey}_{filt}.fits")

    img = fits.open(fits_file)
    img = remove_nan(img)

    header = img[0].header
    data = img[0].data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    # extract galaxy
    gal_obj, _ = extract_objects(
        data_sub,
        bkg_rms,
        host_ra,
        host_dec,
        threshold,
        img_wcs,
        gal_dist_thresh,
    )
    if optimize_kronrad:
        gain = 1  # doesn't matter here
        opt_res = optimize_kron_flux(data_sub, bkg_rms, gain, gal_obj, eps)
        _, _, kronrad, scale = opt_res
    else:
        scale = 2.5
        kronrad, _ = sep.kron_radius(
            data_sub,
            gal_obj["x"],
            gal_obj["y"],
            gal_obj["a"],
            gal_obj["b"],
            gal_obj["theta"],
            6.0,
        )

    if save_plots is True:
        outfile = os.path.join(obj_dir, f"global_{survey}_{filt}.jpg")
        plot_detected_objects(
            data_sub, gal_obj, scale * kronrad, img_wcs, ra, dec, outfile
        )

    if survey == "DES":
        flip = True
    else:
        flip = False

    if save_aperture_params is True:
        outfile = os.path.join(
            obj_dir, f"{survey}_{filt}_aperture_parameters.pickle"
        )
        with open(outfile, "wb") as fp:
            aperture_parameters = gal_obj, img_wcs, kronrad, scale, flip
            pickle.dump(aperture_parameters, fp, protocol=4)

    return gal_obj, img_wcs, kronrad, scale, flip


def photometry(
    name,
    host_ra,
    host_dec,
    filt,
    survey,
    ra=None,
    dec=None,
    bkg_sub=None,
    threshold=10,
    use_mask=True,
    correct_extinction=True,
    aperture_params=None,
    optimize_kronrad=True,
    eps=0.0001,
    gal_dist_thresh=-1,
    save_plots=True,
):
    """Calculates the global aperture photometry of a galaxy using
    the Kron flux.

    **Note:** the galaxy must be ideally centred in the image.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy declination of the galaxy in degrees.
    filt: str
        Filter to use to load the fits file.
    survey: str
        Survey to use for the zero-points and pixel scale.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: bool, default ``None``
        If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: float, default `10`
        Threshold used by `sep.extract()` to extract objects.
    use_mask: bool, default `True`
        If `True`, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: bool, default `True`
        If `True`, corrects for Milky-Way extinction using the recalibrated dust maps
        by Schlafly & Finkbeiner (2011) and the extinction law from Fitzpatrick (1999).
    aperture_params: tuple, default `None`
        Tuple with objects info and Kron parameters. Used for
        common aperture. If given, the Kron parameters are not
        re-calculated
    optimize_kronrad: bool, default `True`
        If `True`, the Kron radius is optimized, increasing the
        aperture size until the flux does not increase.
    eps: float, default ``0.0001``
        The Kron radius is increased until the change in flux is lower than ``eps``.
        A value of 0.0001 means 0.01% change in flux.
    gal_dist_thresh: float, default ``-1``.
        Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    save_plots: bool, default `True`
        If `True`, the mask and galaxy aperture figures are saved.

    Returns
    -------
    mag: float
        Aperture magnitude.
    mag_err: float
        Error on the aperture magnitude.
    flux: float
        Aperture flux.
    total_flux_err: float
        Total flux error on the aperture flux.
    """
    check_survey_validity(survey)
    check_work_dir(workdir)
    obj_dir = os.path.join(workdir, name)
    if use_mask:
        suffix = "masked_"
    else:
        suffix = ""
    fits_file = os.path.join(obj_dir, f"{suffix}{survey}_{filt}.fits")

    img = fits.open(fits_file)
    img = remove_nan(img)

    header = img[0].header
    data = img[0].data
    gain = get_image_gain(header, survey)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    if aperture_params is not None:
        gal_obj, master_img_wcs, kronrad, scale, flip2 = aperture_params

        if survey == "DES":
            flip = True
        else:
            flip = False
        if flip == flip2:
            flip_ = False
        else:
            flip_ = True

        a0 = gal_obj["a"]
        gal_obj = adapt_aperture(gal_obj, master_img_wcs, img_wcs, flip_)
        # factor used for scaling the Kron radius between different pixel scales
        conv_factor = gal_obj["a"] / a0

        flux, flux_err = kron_flux(
            data_sub, bkg_rms, gain, gal_obj, kronrad * conv_factor, scale
        )
        flux, flux_err = flux[0], flux_err[0]
    else:
        # extract galaxy
        gal_obj, _ = extract_objects(
            data_sub,
            bkg_rms,
            host_ra,
            host_dec,
            threshold,
            img_wcs,
            gal_dist_thresh,
        )

        # aperture photometry
        # This uses what would be the default SExtractor parameters.
        # See https://sep.readthedocs.io/en/v1.1.x/apertures.html
        if optimize_kronrad:
            opt_res = optimize_kron_flux(data_sub, bkg_rms, gain, gal_obj, eps)
            flux, flux_err, kronrad, scale = opt_res
        else:
            kronrad, _ = sep.kron_radius(
                data_sub,
                gal_obj["x"],
                gal_obj["y"],
                gal_obj["a"],
                gal_obj["b"],
                gal_obj["theta"],
                6.0,
            )
            scale = 2.5
            flux, flux_err = kron_flux(
                data_sub, bkg_rms, gain, gal_obj, kronrad, scale
            )
            flux, flux_err = flux[0], flux_err[0]

    ap_area = np.pi * gal_obj["a"][0] * gal_obj["b"][0]

    mag, mag_err, flux, flux_err = magnitude_calculation(
        flux,
        flux_err,
        survey,
        filt,
        ap_area,
        header,
        bkg_rms,
    )

    if correct_extinction is True:
        A_ext = calc_extinction(filt, survey, host_ra, host_dec)
        mag -= A_ext

    if save_plots is True:
        outfile = os.path.join(obj_dir, f"global_{survey}_{filt}.jpg")
        plot_detected_objects(
            data_sub, gal_obj, scale * kronrad, img_wcs, ra, dec, outfile
        )

    return mag, mag_err, flux, flux_err


def multi_band_phot(
    name,
    host_ra,
    host_dec,
    filters=None,
    survey="PS1",
    ra=None,
    dec=None,
    bkg_sub=None,
    threshold=10,
    use_mask=True,
    correct_extinction=True,
    aperture_params=None,
    common_aperture=True,
    coadd_filters="riz",
    optimize_kronrad=True,
    eps=0.0001,
    gal_dist_thresh=-1,
    save_plots=True,
    save_results=True,
    raise_exception=False,
):
    """Calculates multi-band aperture photometry of the host galaxy
    for an object.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy declination of the galaxy in degrees.
    filters: str, default, ``None``
        Filters to use to load the fits files. If `None` use all
        the filters of the given survey.
    survey: str, default ``PS1``
        Survey to use for the zero-points and pixel scale.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: bool, default ``None``
        If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: float, default ``10``
        Threshold used by :func:`sep.extract()` to extract objects.
    use_mask: bool, default ``True``
        If ``True``, the masked fits files are used. These must have
        been created beforehand.
    correct_extinction: bool, default ``True``
        If ``True``, the magnitudes are corrected for extinction.
    aperture_params: tuple, default `None`
        Tuple with objects info and Kron parameters. Used for common aperture. If given,
        the Kron parameters are not re-calculated. If given, this supersedes the use of
        coadds for common aperture (``common_aperture`` parameter).
    common_aperture: bool, default ``True``
        If ``True``, use a coadd image for common aperture photometry. This is not used
        if ``aperture_params`` is given.
    coadd_filters: str, default ``riz``
        Filters of the coadd image. Used for common aperture photometry.
    optimize_kronrad: bool, default ``True``
        If ``True``, the Kron radius is optimized, increasing the
        aperture size until the flux does not increase.
    eps: float, default ``0.0001``
        The Kron radius is increased until the change in flux is lower than ``eps``.
        A value of 0.0001 means 0.01% change in flux.
        when optimizing the Kron radius.
    gal_dist_thresh: float, default ``-1``.
        Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    save_plots: bool, default ``True``
        If ``True``, the mask and galaxy aperture figures are saved.
    save_results: bool, default ``True``
        If ``True``, the magnitudes are saved into a csv file.
    raise_exception: bool, default ``False``
        If ``True``, an exception is raised if the photometry fails for any filter.

    Returns
    -------
    results_dict: dict
        Dictionary with the object's photometry and other info.

    Examples
    --------
    >>> import hostphot.global_photometry as gp
    >>> name = 'SN2004eo'
    >>> host_ra, host_dec = 308.2092, 9.92755  # coords of host galaxy of SN2004eo
    >>> ra, dec =  308.22579, 9.92853 # coords of SN2004eo
    >>> results = gp.multi_band_phot(name, host_ra, host_dec,
                            survey=survey, ra=ra, dec=dec,
                            use_mask=True, common_aperture=True,
                            coadd_filters='riz', save_plots=True)
    """
    check_survey_validity(survey)
    if filters is None:
        if survey=='HST':
            raise ValueError("For HST, the filter needs to be specified!")
        filters = get_survey_filters(survey)
    else:
        check_filters_validity(filters, survey)
    if survey=='HST':
        filters = [filters]

    results_dict = {
        "name": name,
        "host_ra": host_ra,
        "host_dec": host_dec,
        "survey": survey,
    }

    if common_aperture is True and aperture_params is None:
        aperture_params = extract_kronparams(
            name,
            host_ra,
            host_dec,
            coadd_filters,
            survey,
            ra=ra,
            dec=dec,
            bkg_sub=bkg_sub,
            threshold=threshold,
            use_mask=use_mask,
            optimize_kronrad=optimize_kronrad,
            eps=eps,
            gal_dist_thresh=gal_dist_thresh,
            save_plots=save_plots,
        )

    for filt in filters:
        try:
            mag, mag_err, flux, flux_err = photometry(
                name,
                host_ra,
                host_dec,
                filt,
                survey,
                ra=ra,
                dec=dec,
                bkg_sub=bkg_sub,
                threshold=threshold,
                use_mask=use_mask,
                correct_extinction=correct_extinction,
                aperture_params=aperture_params,
                optimize_kronrad=optimize_kronrad,
                eps=eps,
                gal_dist_thresh=gal_dist_thresh,
                save_plots=save_plots,
            )
            results_dict[filt] = mag
            results_dict[f"{filt}_err"] = mag_err
            results_dict[f"{filt}_flux"] = flux
            results_dict[f"{filt}_flux_err"] = flux_err
        except Exception as exc:
            if raise_exception is True:
                raise Exception(f"{filt}-band: {exc}")
            else:
                results_dict[filt] = np.nan
                results_dict[f"{filt}_err"] = np.nan

    if save_results is True:
        outfile = os.path.join(workdir, name, f"{survey}_global.csv")
        phot_df = pd.DataFrame(
            {key: [val] for key, val in results_dict.items()}
        )
        phot_df.to_csv(outfile, index=False)

    return results_dict



import os
import pickle
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

import sep
from astropy.io import fits
from astropy import wcs, units as u
from astropy.coordinates import SkyCoord, concatenate
from astropy.convolution import (
    Gaussian2DKernel,
    convolve_fft,
    interpolate_replace_nans,
)

from hostphot._constants import workdir
from hostphot.image_cleaning import remove_nan
from hostphot.objects_detect import (
    extract_objects,
    find_gaia_objects,
    find_catalog_objects,
    cross_match,
)
from hostphot.utils import (
    check_survey_validity,
    pixel2pixel,
    update_axislabels,
    survey_pixel_scale,
    bkg_surveys,
    adapt_aperture,
)

import warnings
from astropy.utils.exceptions import AstropyWarning


# ----------------------------------------
def mask_image(data, objects, r=5, sigma=20):
    """Masks objects in an image (2D array) by convolving it with
    a 2D Gaussian kernel.

    Parameters
    ----------
    data: ndarray
        Image data.
    objects: array
        Objects extracted with :func:`sep.extract()`.
    r: float, default ``5``
        Scale of the semi-mayor and semi-minor axes
        of the ellipse of the `objects`.
    sigma: float, default ``20``
        Standard deviation in pixel units of the 2D Gaussian kernel
        used to convolve the image.

    Returns
    -------
    masked_data: 2D array
        Masked image data.
    """
    mask = np.zeros(data.shape, dtype=bool)
    sep.mask_ellipse(
        mask,
        objects["x"],
        objects["y"],
        objects["a"],
        objects["b"],
        objects["theta"],
        r=r,
    )

    masked_data = data.copy()
    masked_data[mask] = np.nan
    # mask data by convolving it with a 2D Gaussian kernel
    # with the same sigma in x and y
    kernel = Gaussian2DKernel(sigma)
    masked_data = interpolate_replace_nans(masked_data, kernel, convolve_fft)

    return masked_data


def create_mask(
    name,
    host_ra,
    host_dec,
    filt,
    survey,
    ra=None,
    dec=None,
    bkg_sub=None,
    threshold=15,
    sigma=8,
    crossmatch=False,
    gal_dist_thresh=-1,
    extract_params=False,
    common_params=None,
    save_plots=True,
    save_mask_params=True,
):
    """Calculates the aperture parameters for common aperture.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension in degrees.
    host_dec: float
        Host-galaxy declination in degrees.
    filt: str or list
        Filter to use to load the fits file. List is commonly used for coadds.
    survey: str
        Survey to use for the zero-points and correct filter path.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting.
    bkg_sub: bool, default ``None``
        If ``True``, the image gets background subtracted. By default, only
        the images that need it get background subtracted (WISE, 2MASS and
        VISTA).
    threshold: float, default `15`
        Threshold used by :func:`sep.extract()` to extract objects.
    sigma: float, default ``8``
        Standard deviation in pixel units of the 2D Gaussian kernel
        used to convolve the image.
    crossmatch: bool, default ``False``
        If ``True``, the detected objects are cross-matched with a
        Gaia catalog.
    gal_dist_thresh: float, default ``-1``.
        Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).
    extract_params: bool, default ``False``
        If ``True``, returns the parameters listed below.
    common_params: tuple, default ``None``
        Parameters to use for common masking of different filters.
        These are the same as the outputs of this function.
    save_plots: bool, default ``True``
        If ``True``, the mask and galaxy aperture figures are saved.
    save_mask_params: bool, default `True`
        If `True`, the extracted mask parameters are saved into a pickle file.

    Returns
    -------
    **This are only returned if ``extract_params==True``.**
    gal_obj: array
        Galaxy object.
    nogal_obj: array
        Non-galaxy objects.
    img_wcs: WCS
        Image's WCS.
    flip: bool
        Whether to flip the orientation of the
        aperture. Only used for DES images.
    """
    check_survey_validity(survey)
    if isinstance(filt, list):
        filt = "".join(f for f in filt)

    obj_dir = os.path.join(workdir, name)
    fits_file = os.path.join(obj_dir, f"{survey}_{filt}.fits")
    img = fits.open(fits_file)
    img = remove_nan(img)

    header = img[0].header
    data = img[0].data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if (bkg_sub is None and survey in bkg_surveys) or bkg_sub is True:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    if common_params is None:
        # extract objects
        gal_obj, nogal_objs = extract_objects(
            data_sub,
            bkg_rms,
            host_ra,
            host_dec,
            threshold,
            img_wcs,
            gal_dist_thresh,
        )
        # preprocessing: cross-match extracted objects with a catalog
        # using two Gaia catalogs as they do not always include the
        # same objects.
        if crossmatch:
            cat_coord1 = find_gaia_objects(host_ra, host_dec)
            cat_coord2 = find_catalog_objects(host_ra, host_dec)
            cat_coord = concatenate([cat_coord1, cat_coord2])
            nogal_objs = cross_match(nogal_objs, img_wcs, cat_coord)
    else:
        # use objects previously extracted
        # the aperture/ellipse parameters are updated accordingly
        gal_obj, nogal_objs, master_img_wcs, flip2 = common_params

        if survey == "DES":
            flip = True
        else:
            flip = False
        if flip == flip2:
            flip_ = False
        else:
            flip_ = True

        gal_obj = adapt_aperture(gal_obj, master_img_wcs, img_wcs, flip_)
        nogal_objs = adapt_aperture(nogal_objs, master_img_wcs, img_wcs, flip_)

    masked_data = mask_image(data_sub, nogal_objs, sigma=sigma)
    img[0].data = masked_data
    outfile = os.path.join(obj_dir, f"masked_{survey}_{filt}.fits")
    img.writeto(outfile, overwrite=True)

    if save_plots:
        outfile = os.path.join(obj_dir, f"masked_{survey}_{filt}.jpg")
        plot_masked_image(
            data_sub,
            masked_data,
            nogal_objs,
            img_wcs,
            gal_obj,
            host_ra,
            host_dec,
            ra,
            dec,
            outfile,
        )

    if survey == "DES":
        flip = True
    else:
        flip = False

    if save_mask_params is True:
        outfile = os.path.join(
            obj_dir, f"{survey}_{filt}_mask_parameters.pickle"
        )
        with open(outfile, "wb") as fp:
            mask_parameters = gal_obj, nogal_objs, img_wcs, flip
            pickle.dump(mask_parameters, fp, protocol=4)

    if extract_params:
        return gal_obj, nogal_objs, img_wcs, flip


def plot_masked_image(
    data,
    masked_data,
    objects,
    img_wcs,
    gal_obj=None,
    host_ra=None,
    host_dec=None,
    ra=None,
    dec=None,
    outfile=None,
):
    """Plots the masked image together with the original image and
    the detected objects.

    Parameters
    ----------
    data: ndarray
        Image data.
    masked_data: 2D array
         Masked image data.
    objects: array
        Objects extracted with :func:`sep.extract()`.
    img_wcs: WCS
        Image's WCS.
    gal_obj: array, default ``None``
        Galaxy object.
    host_ra: float, default ``None``
       Right ascension of the galaxy, in degrees. Used for plotting the position of the galaxy.
    host_dec: float, default ``None``
       Declination of the galaxy, in degrees. Used for plotting the position of the galaxy.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    outfile: str, default ``None``
        If given, path where to save the output figure.
    """
    r = 4  # scale
    m, s = np.nanmean(data), np.nanstd(data)

    fig = plt.figure(figsize=(24, 10))
    ax0 = plt.subplot(131, projection=img_wcs)
    ax1 = plt.subplot(132, projection=img_wcs)
    ax2 = plt.subplot(133, projection=img_wcs)
    axes = [ax0, ax1, ax2]

    for ax in axes:
        update_axislabels(ax)
        ax.imshow(
            data,
            interpolation="nearest",
            cmap="gray",
            vmin=m - s,
            vmax=m + s,
            origin="lower",
        )
    for ax in axes[1:]:
        ax.coords[1].set_axislabel("")
        ax.coords[1].set_ticklabel_visible(False)

    for i in range(len(objects)):
        e = Ellipse(
            xy=(objects["x"][i], objects["y"][i]),
            width=2.5 * r * objects["a"][i],
            height=2.5 * r * objects["b"][i],
            angle=objects["theta"][i] * 180.0 / np.pi,
        )
        e.set_facecolor("none")
        e.set_edgecolor("red")
        e.set_linewidth(1.5)
        ax1.add_artist(e)

    # plot galaxy pseudo-aperture
    if gal_obj is not None:
        e = Ellipse(
            xy=(gal_obj["x"][0], gal_obj["y"][0]),
            width=2.5 * r * gal_obj["a"][0],
            height=2.5 * r * gal_obj["b"][0],
            angle=gal_obj["theta"][0] * 180.0 / np.pi,
        )
        e.set_facecolor("none")
        e.set_edgecolor("red")
        e.set_linestyle("dotted")
        e.set_linewidth(3)
        ax1.add_artist(e)

        ax1.scatter(
            gal_obj["x"][0],
            gal_obj["y"][0],
            marker="x",
            s=140,
            c="r",
            label="Identified galaxy center",
        )

    ax2.imshow(
        masked_data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )

    ax0.set_title("Initial Image", fontsize=24)
    ax1.set_title("Detected Objects", fontsize=24)
    ax2.set_title("Masked Image", fontsize=24)

    # plot SN position
    if (host_ra is not None) and (host_dec is not None):
        coord = SkyCoord(
            ra=host_ra, dec=host_dec, unit=(u.degree, u.degree), frame="icrs"
        )
        x, y = img_wcs.world_to_pixel(coord)
        for ax in axes[1:]:
            ax.scatter(
                x,
                y,
                marker="P",
                s=140,
                c="r",
                edgecolor="gold",
                label="Galaxy position",
            )

    # plot SN position
    if (ra is not None) and (dec is not None):
        coord = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
        x, y = img_wcs.world_to_pixel(coord)
        for ax in axes[1:]:
            ax.scatter(
                x,
                y,
                marker="*",
                s=200,
                c="g",
                edgecolor="gold",
                label="SN position",
            )

    axes[1].legend(ncol=2, fontsize=14)
    if outfile:
        basename = os.path.basename(outfile)
        title = os.path.splitext(basename)[0]
        title = "-".join(part for part in title.split("_"))
        fig.suptitle(title, fontsize=28)
        plt.tight_layout()
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.show()



import unittest
from hostphot.cutouts import download_images
from hostphot.coadd import coadd_images
from hostphot.image_masking import create_mask


class TestHostPhot(unittest.TestCase):
    def test_preprocessing(self):
        coadd_filters = "riz"
        survey = "PS1"
        sn_name = "2002fk"
        host_ra = 50.527333
        host_dec = -15.400056

        download_images(sn_name, host_ra, host_dec, survey=survey)

        # coadd
        coadd_images(sn_name, coadd_filters, survey)

        # masking
        coadd_mask_params = create_mask(
            sn_name,
            host_ra,
            host_dec,
            filt=coadd_filters,
            survey=survey,
            extract_params=True,
            crossmatch=True,
        )

        # apply mask on single-filter images
        for filt in "grizy":
            create_mask(
                sn_name,
                host_ra,
                host_dec,
                filt,
                survey=survey,
                common_params=coadd_mask_params,
            )


if __name__ == "__main__":
    unittest.main()



import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

import sep
from astroquery.gaia import Gaia
from astroquery.mast import Catalogs
from astropy import units as u
from astropy.coordinates import SkyCoord

from hostphot.utils import update_axislabels


def extract_objects(
    data, err, host_ra, host_dec, threshold, img_wcs, dist_thresh=-1
):
    """Extracts objects and their ellipse parameters. The function :func:`sep.extract()`
    is used.

    If there is no detected object within a distance of ``dist_thresh`` from the galaxy
    coordinates, it means that the galaxy was not correctly identified.

    Parameters
    ----------
    data: ndarray
        Image data.
    err: 2D array
        Background error of the image.
    host_ra: float
        Host-galaxy Right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy Declination of the galaxy in degrees.
    threshold: float
        Source with flux above ``threshold*bkg_rms`` are extracted.
        See :func:`sep.extract()` for more information.
    img_wcs: WCS
        Image's WCS.
    pixel_scale: float
        Pixel scale, in units of arcsec/pixel, used to convert from pixel units
        to arcseconds.
    dist_thresh: float, default ``-1``.
        Distance in arcsec to crossmatch the galaxy coordinates with a detected object,
        where the object nearest to the galaxy position is considered as the galaxy (within
        the given threshold). If no objects are found within the given distance threshold,
        the galaxy is considered as not found and a warning is printed. If a non-positive value
        is given, the threshold is considered as infinite, i.e. the closest detected object is
        considered as the galaxy (default option).

    Returns
    -------
    gal_obj: numpy array
        Galaxy object extracted.
    nogal_objs: numpy array
        All objects extracted except for the galaxy.
    """
    # extract objects with Source Extractor
    objects = sep.extract(data, threshold, err=err)

    objs_coords = img_wcs.pixel_to_world(objects["x"], objects["y"])
    objs_ra, objs_dec = objs_coords.ra.value, objs_coords.dec.value
    dist = np.sqrt(
        (objs_ra - host_ra) ** 2 * (np.cos(host_dec * np.pi / 180) ** 2)
        + (objs_dec - host_dec) ** 2
    )  # in degrees
    dist_arcsec = dist * 3600

    if dist_thresh <= 0.0:
        dist_thresh = np.inf

    if any(dist_arcsec <= dist_thresh):
        gal_id = np.argmin(dist_arcsec)
        gal_obj = objects[gal_id : gal_id + 1]
    else:
        gal_obj = None
        gal_id = -99
        print("WARNING: the galaxy was no detected")

    objs_id = [i for i in range(len(objects)) if i != gal_id]
    nogal_objs = objects.take(objs_id)

    return gal_obj, nogal_objs


def find_gaia_objects(ra, dec, rad=0.15):
    """Finds objects using the Gaia DR3 catalog for the given
    coordinates in a given radius.

    Parameters
    ----------
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    rad: float, default ``0.15``
        Search radius in degrees.

    Returns
    -------
    gaia_coord: SkyCoord object
        Coordinates of the objects found.
    """
    Gaia.MAIN_GAIA_TABLE = "gaiaedr3.gaia_source"
    Gaia.ROW_LIMIT = -1
    coord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")
    width = u.Quantity(rad, u.deg)
    height = u.Quantity(rad, u.deg)
    try:
        gaia_cat = Gaia.query_object_async(
            coordinate=coord, width=width, height=height
        )
    except Exception as exc:
        print(exc)
        print("No objects found with Gaia DR3, switching to DR2")
        Gaia.MAIN_GAIA_TABLE = "gaiadr2.gaia_source"
        gaia_cat = Gaia.query_object_async(
            coordinate=coord, width=width, height=height
        )

    gaia_ra = np.array(gaia_cat["ra"].value)
    gaia_dec = np.array(gaia_cat["dec"].value)
    gaia_coord = SkyCoord(
        ra=gaia_ra, dec=gaia_dec, unit=(u.degree, u.degree), frame="icrs"
    )

    return gaia_coord


def find_catalog_objects(ra, dec, rad=0.15):
    """Finds objects using the TESS image cutouts (Tic) catalog
    for the given coordinates in a given radius.

    **Note:** this catalog includes objects from these sources:
        HIP, TYC, UCAC, TWOMASS, SDSS, ALLWISE, GAIA, APASS, KIC

    Parameters
    ----------
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    rad: float, default ``0.15``
        Search radius in degrees.

    Returns
    -------
    cat_coord: SkyCoord object
        Coordinates of the objects found.
    """
    coord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")
    cat_data = Catalogs.query_criteria(
        catalog="Tic", radius=rad, coordinates=coord, objType="STAR"
    )

    cat_ra = np.array(cat_data["ra"].value)
    cat_dec = np.array(cat_data["dec"].value)
    cat_coord = SkyCoord(
        ra=cat_ra, dec=cat_dec, unit=(u.degree, u.degree), frame="icrs"
    )

    return cat_coord


def cross_match(objects, img_wcs, coord, dist_thresh=1.0):
    """Cross-matches objects with a given set of coordinates.
    Those with a distance of less than ``dist_thresh`` are selected.

    Parameters
    ----------
    objects: array
        Objects detected with :func:`sep.extract()`.
    img_wcs: WCS
        WCS of the image from which the objects were extracted.
    coord: SkyCoord object
        Coordinates for the cross-match.
    dist_thresh: float, default ``1.0``
        Distance in arcsec to crossmatch the objects with
        the given coordinates.
    """
    # coordinates in arcsec
    objs_coord = img_wcs.pixel_to_world(objects["x"], objects["y"])
    objs_ra = objs_coord.ra.to(u.arcsec).value
    objs_dec = objs_coord.dec.to(u.arcsec).value

    cat_ra = np.array(coord.ra.to(u.arcsec).value)
    cat_dec = np.array(coord.dec.to(u.arcsec).value)

    objs_id = []
    for i in range(len(objects)):
        ra_dist = objs_ra[i] - cat_ra
        dec_dist = objs_dec[i] - cat_dec
        dist_arcsec = np.sqrt(ra_dist**2 + dec_dist**2)
        if any(dist_arcsec <= dist_thresh):
            objs_id.append(i)

    objs = objects.take(objs_id)

    return objs


def plot_detected_objects(
    data, objects, scale, img_wcs, ra=None, dec=None, outfile=None
):
    """Plots the objects extracted with :func:`sep.extract()``.

    Parameters
    ----------
    data: ndarray
        Data of an image.
    objects: array
        Objects detected with :func:`sep.extract()`.
    scale: float
        Scale of the ellipse's semi-mayor and semi-minor axes.
    img_wcs: WCS
        Image's WCS.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    outfile: str, default ``None``
        If given, path where to save the output figure.
    """
    m, s = np.nanmean(data), np.nanstd(data)

    fig = plt.figure(figsize=(10, 10))
    ax = plt.subplot(projection=img_wcs)
    update_axislabels(ax)

    im = ax.imshow(
        data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )

    e = Ellipse(
        xy=(objects["x"][0], objects["y"][0]),
        width=2 * scale * objects["a"][0],
        height=2 * scale * objects["b"][0],
        angle=objects["theta"][0] * 180.0 / np.pi,
    )

    e.set_facecolor("none")
    e.set_edgecolor("red")
    e.set_linewidth(1.5)
    ax.add_artist(e)

    # plot SN position
    if (ra is not None) and (dec is not None):
        coord = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
        px, py = img_wcs.world_to_pixel(coord)
        ax.scatter(px, py, marker="*", s=200, c="g", edgecolor="gold")

    if outfile:
        basename = os.path.basename(outfile)
        title = os.path.splitext(basename)[0]
        title = "-".join(part for part in title.split("_"))
        fig.suptitle(title, fontsize=28)
        plt.tight_layout()
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.show()



import unittest
import warnings
from hostphot.cutouts import download_images


class TestHostPhot(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestHostPhot, self).__init__(*args, **kwargs)
        # object used for most surveys
        self.sn_name = "2002fk"
        self.ra = 50.527333
        self.dec = -15.400056

    def test_cutouts_PS1(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="PS1"
        )

    def test_cutouts_DES(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="DES"
        )

    def test_cutouts_SDSS(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="SDSS"
        )

    def test_cutouts_GALEX(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="GALEX"
        )

    def test_cutouts_WISE(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="WISE"
        )

    def test_cutouts_2MASS(self):
        download_images(
            self.sn_name, self.ra, self.dec, overwrite=True, survey="2MASS"
        )

    def test_cutouts_unWISE(self):
        for version in ["neo1", "neo2", "allwise"]:
            download_images(
                self.sn_name,
                self.ra,
                self.dec,
                overwrite=True,
                survey="unWISE",
                version=version,
            )

    def test_cutouts_LegacySurvey(self):
        download_images(
            self.sn_name,
            self.ra,
            self.dec,
            overwrite=True,
            survey="LegacySurvey",
        )

    def test_cutouts_Spitzer(self):
        name = "Spitzer_test"
        ra, dec = 52.158591, -27.891113
        download_images(name, ra, dec, overwrite=True, survey="Spitzer")

    def test_cutouts_VISTA(self):
        name = "VISTA_test"
        # use different coordinates for each survey as they don't overlap
        surveys = {
            "VHS": [120, -60],
            "VIDEO": [36.1, -5],
            "VIKING": [220.5, 0.0],
        }

        try:
            for version, coords in surveys.items():
                ra, dec = coords
                download_images(
                    name,
                    ra,
                    dec,
                    overwrite=True,
                    survey="VISTA",
                    version=version,
                )
        except Exception as exc:
            warnings.warn(
                "The VISTA SCIENCE ARCHIVE might be having issues..."
            )
            print("Skipping this test...")
            print(exc)


if __name__ == "__main__":
    unittest.main()



import os

from astropy.io import fits
from reproject import reproject_interp
from reproject.mosaicking import reproject_and_coadd

from hostphot._constants import workdir

import warnings
from astropy.utils.exceptions import AstropyWarning


# ----------------------------------------


def coadd_images(name, filters="riz", survey="PS1"):
    """Reprojects and coadds images for the choosen filters for
    common-aperture photometry.

    Parameters
    ----------
    name: str
        Name to be used for finding the images locally.
    filters: str or list, default ``riz``
        Filters to use for the coadd image.
    survey: str, default ``PS1``
        Survey to use as prefix for the images.

    Examples
    --------
    >>> from hostphot.coadd import coadd_images
    >>> name = 'SN2004eo'
    >>> survey = 'PS1'
    >>> coadd_filters = 'riz'
    >>> coadd_images(name, filters=coadd_filters, survey=survey)  # creates a new fits file
    """
    obj_dir = os.path.join(workdir, name)
    fits_files = [
        os.path.join(obj_dir, f"{survey}_{filt}.fits") for filt in filters
    ]

    hdu_list = []
    for fits_file in fits_files:
        fits_image = fits.open(fits_file)
        hdu_list.append(fits_image[0])

    hdu_list = fits.HDUList(hdu_list)
    # use the last image as reference
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coadd = reproject_and_coadd(
            hdu_list, fits_image[0].header, reproject_function=reproject_interp
        )
    fits_image[0].data = coadd[0]
    if isinstance(filters, list):
        filters = "".join(filt for filt in filters)
    outfile = os.path.join(obj_dir, f"{survey}_{filters}.fits")
    fits_image.writeto(outfile, overwrite=True)



import unittest
import numpy as np
from hostphot.cutouts import download_images
import hostphot.local_photometry as lp


class TestHostPhot(unittest.TestCase):
    def test_local_phot(self):
        sn_name = "2002fk"
        ra = 50.52379
        dec = -15.40089
        z = 0.007482
        survey = "PS1"

        ap_radii = [3, 4]  # in units of kpc
        download_images(sn_name, ra, dec, survey=survey)
        phot = lp.multi_band_phot(
            sn_name,
            ra,
            dec,
            z,
            survey="PS1",
            filters="grizy",
            ap_radii=ap_radii,
            use_mask=False,
            save_plots=True,
            raise_exception=True
        )
        mags = [phot[filt] for filt in ["g_4", "r_4", "i_4", "z_4"]]
        # pre-calculated magnitudes
        ref_mags = [12.26, 11.89, 11.72, 11.57]

        err_msg = (
            "Large difference between calculated and reference magnitudes"
        )
        np.testing.assert_allclose(mags, ref_mags, rtol=0.03, err_msg=err_msg)


if __name__ == "__main__":
    unittest.main()



import os
import glob
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from astropy import wcs
from astropy.io import fits
from astropy.stats import sigma_clipped_stats
from photutils.utils import calc_total_error
from photutils.aperture import EllipticalAperture

import warnings
from astropy.utils.exceptions import AstropyWarning

import hostphot

hostphot_path = hostphot.__path__[0]
config_file = os.path.join(hostphot_path, "filters", "config.txt")
config_df = pd.read_csv(config_file, delim_whitespace=True)

# surveys that need background subtraction
bkg_surveys = ['2MASS', 'WISE', 'VISTA']

def calc_sky_unc(image, exptime):
    """Calculates the uncertainty of the image from the
    sky standard deviation, sigma-clipped STD.

    Parameters
    ----------
    image: ndarray
        Image in a 2D numpy array.
    exptime: float
        Exposure time of the image.

    Returns
    -------
    error: float
        Estimated error of the image.
    """
    mask = image != 0
    avg, sky, sky_std = sigma_clipped_stats(image[mask], sigma=3.0)
    error = calc_total_error(image, sky_std, exptime)

    return error


def pixel2pixel(x1, y1, img_wcs1, img_wcs2):
    """Convert the pixel coordinates from one image to another."""
    coord1 = img_wcs1.pixel_to_world(x1, y1)
    x2, y2 = img_wcs2.world_to_pixel(coord1)

    return x2, y2


def check_survey_validity(survey):
    """Check whether the given survey is whithin the valid
    options.

    Parameters
    ----------
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.
    """
    global config_df
    surveys = list(config_df.survey)
    assert survey in surveys, f"Survey '{survey}' not in {surveys}"


def get_survey_filters(survey):
    """Gets all the valid filters for the given survey.

    Parameters
    ----------
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    filters: str
        Filters for the given survey.
    """
    check_survey_validity(survey)

    if survey=='HST':
        # For HST, the filter needs to be specified
        return None

    global config_df
    survey_df = config_df[config_df.survey == survey]
    filters = survey_df.filters.values[0]

    if "," in filters:
        filters = filters.split(",")

    return filters


def survey_zp(survey):
    """Returns the zero-point for a given survey.

    **Note:** for ``PS1``, an extra :math:`+2.5*np.log10(exptime)`
    needs to be added afterwards.

    Parameters
    ----------
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    zp_dict: dict or str
        Zero-points for all the filters in the given survey. If the survey zero-point
        is different for each image, the string ``header`` is returned.
    """
    check_survey_validity(survey)
    filters = get_survey_filters(survey)

    global config_df
    survey_df = config_df[config_df.survey == survey]
    zps = survey_df.zp.values[0]

    if zps == "header":
        return zps

    if "," in zps:
        zps = zps.split(",")
        zp_dict = {filt: float(zp) for filt, zp in zip(filters, zps)}
    elif survey=='SDSS':
        # SDSS zero-points are not exactly in AB:
        # https://www.sdss4.org/dr12/algorithms/fluxcal/#SDSStoAB
        zp_dict = {filt: float(zps) for filt in filters}
        zp_dict['u'] -= 0.04
        zp_dict['z'] += 0.02
    else:
        zp_dict = {filt: float(zps) for filt in filters}

    return zp_dict


def get_image_gain(header, survey):
    """Returns the gain from an image's header.

    **Note:** for ``SDSS`` this is assumed to be one
    as it should already be included.

    Parameters
    ----------
    header: fits header
        Header of an image.
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    gain: float
        Gain value.
    """
    check_survey_validity(survey)
    if survey == "PS1":
        gain = header["HIERARCH CELL.GAIN"]
    elif survey == "DES":
        gain = header["GAIN"]
    elif survey == "SDSS":
        gain = 1.0
    elif survey == "2MASS":
        # the value comes from https://wise2.ipac.caltech.edu/staff/jarrett/2mass/3chan/noise/
        # Note that it is different from the value of
        # https://iopscience.iop.org/article/10.1086/498708/pdf (8 e ADU^-1)
        gain = 10.0
    elif survey == "LegacySurvey":
        gain = 30  # assumed similar to DES
    elif survey == "Spitzer":
        # also in the "EXPGAIN" keyword
        if header["INSTRUME"] == "IRAC":
            # Table 2.4 of IRAC Instrument Handbook
            gain = 3.8  # all filters have similar gain
            gain *= header["EFCONV"]  # convert DN/s to MJy/sr
        elif header["INSTRUME"] == "MIPS":
            # Table 2.4 of MIPS Instrument Handbook
            gain = 5.0
    elif survey == "VISTA":
        # use median from http://casu.ast.cam.ac.uk/surveys-projects/vista/technical/vista-gain
        gain = 4.19
    elif survey=='HST':
        gain = header['CCDGAIN']
    else:
        gain = 1.0

    return gain


def get_image_readnoise(header, survey):
    """Returns the read noise from an image's header.
    All values are per-pixel values.

    **Note:** for ``SDSS`` this is assumed to be zero
    as it should already be included.

    Parameters
    ----------
    header: fits header
        Header of an image.
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    readnoise: float
        Read noise value.
    """
    check_survey_validity(survey)
    if survey == "PS1":
        readnoise = header["HIERARCH CELL.READNOISE"]
    elif survey == "DES":
        # see https://arxiv.org/pdf/0810.3600.pdf
        readnoise = 7.0  # electrons per pixel
    elif survey == "SDSS":
        readnoise = 0.0
    elif survey == "2MASS":
        # https://iopscience.iop.org/article/10.1086/498708/pdf
        # 6 combined images
        readnoise = 4.5 * np.sqrt(6)  # not used
    elif survey == "LegacySurvey":
        readnoise = 7.0  # assumed similar to DES
    elif survey == "Spitzer":
        if header["INSTRUME"] == "IRAC":
            # Table 2.3 of IRAC Instrument Handbook
            # very rough average
            readnoise_dict = {1: 16.0, 2: 12.0, 3: 10.0, 4: 8.0}
            channel = header["CHNLNUM"]
            readnoise = readnoise_dict[channel]
        elif header["INSTRUME"] == "MIPS":
            # Table 2.4 of MIPS Instrument Handbook
            readnoise = 40.0
    elif survey == "VISTA":
        # very rough average for all filters in
        # http://casu.ast.cam.ac.uk/surveys-projects/vista/technical/vista-gain
        readnoise = 24.0
    elif survey=='HST':
        # tipically 0.0
        readnoise = header['PCTERNOI']
    else:
        readnoise = 0.0

    return readnoise


def get_image_exptime(header, survey):
    """Returns the exposure time from an image's header.

    Parameters
    ----------
    header: fits header
        Header of an image.
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.

    Returns
    -------
    exptime: float
        Exposure time in seconds.
    """
    check_survey_validity(survey)
    if survey in ["PS1", "DES", "GALEX", "VISTA", "Spitzer", "HST"]:
        exptime = float(header["EXPTIME"])
    elif survey == "WISE":
        # see: https://wise2.ipac.caltech.edu/docs/release/allsky/
        # and https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec1_1.html
        if header["BAND"] in [1, 2]:
            exptime = 7.7
        elif header["BAND"] in [3, 4]:
            exptime = 8.8
    elif survey == "2MASS":
        # https://iopscience.iop.org/article/10.1086/498708/pdf
        exptime = 7.8
    elif survey == "LegacySurvey":
        exptime = 900.0  # assumed similar to DES
    else:
        exptime = 1.0

    return exptime


def correct_HST_aperture(filt, ap_area, header):
    """Get the aperture correction for the given configuration.

    Parameters
    ----------
    filt : str
        HST filter, e.g. ``WFC3_UVIS_F225W``.
    ap_area: float
        Aperture area.
    header: fits header
        Header of an image.

    Returns
    -------
    correction: float
        Aperture correction (encircled energy fraction).
    """
    # split instrument and filter
    filt_split = filt.split('_')
    filt = filt_split[-1]
    instrument = filt_split[-2]

    if instrument=='UVIS':
        instrument = header['APERTURE']

    # assuming circular aperture
    # for an ellipse, this would take the average of the axes
    ap_radius = np.sqrt(ap_area/np.pi)

    # get correction curve
    ac_files = glob.glob(os.path.join(hostphot_path, 'filters/HST/*'))
    ac_file = [file for file in ac_files 
               if f'{instrument.lower()}_aper' in file][0]
    ac_df = pd.read_csv(ac_file)

    # linear interpolation of the aperture correction
    apertures = np.array([float(col.replace('APER#', '')) for col in ac_df.columns 
                          if col.startswith('AP')])
    ap_corr = ac_df[ac_df.FILTER==filt].values[0][2:].astype(float)

    cont_apertures = np.arange(0, 9, 0.01)
    cont_ap_corr = np.interp(cont_apertures, apertures, ap_corr)

    # get the closest value
    ind = np.argmin(np.abs(cont_apertures-ap_radius))
    correction = cont_ap_corr[ind]

    return correction

def magnitude_calculation(
    flux,
    flux_err,
    survey,
    filt=None,
    ap_area=0.0,
    header=None,
    bkg_rms=0.0,
):
    """Calculates the calibrated magnitudes and errors.

    Parameters
    ----------
    flux: float
        Aperture flux.
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.
    filt: str, default ``None``
        Survey-specific filter.
    ap_area: float, default ``0.0``
        Aperture area.
    header: fits header, default ``None``
        Header of an image.
    bkg_rms: float, default ``0.0``
        Background noise.

    Returns
    -------
    mag: float
        Apparent magnitude.
    mag_err: float
        Apparent magnitude uncertainty.
    flux: float
        Total flux.
    flux_error: float
        Total uncertainty in flux.
    """
    # get zeropoint
    zp_dict = survey_zp(survey)
    if zp_dict == "header":
        zp = header["MAGZP"]
    else:
        zp = zp_dict[filt]
    if survey == "PS1":
        exptime = get_image_exptime(header, survey)
        zp += 2.5 * np.log10(exptime)

    # error propagation
    mag_err, flux_err = uncertainty_calculation(
        flux,
        flux_err,
        survey,
        filt,
        ap_area,
        header,
        bkg_rms,
    )

    if survey == 'HST':
        # HST needs and aperture correction for the flux
        # see, e.g. https://www.stsci.edu/hst/instrumentation/acs/data-analysis/aperture-corrections
        ap_corr = correct_HST_aperture(filt, ap_area, header)
        flux = flux*ap_corr

    mag = -2.5 * np.log10(flux) + zp

    return mag, mag_err, flux, flux_err


def get_HST_err(filt, header):
    """Obtaines the error propagation from the zeropoint.

    Parameters
    ----------
    filt : str
        HST filter, e.g. ``WFC3_UVIS_F225W``.
    header: fits header
        Header of an image.

    Returns
    -------
    flux_err: float
        Error on PHOTFLAM.
    mag_err: float
        Magnitude error on PHOTFLAM.
    """
    # split instrument and filter
    filt_split = filt.split('_')
    filt = filt_split[-1]
    instrument = filt_split[-2]

    if instrument=='UVIS':
        # APERTURE usually point to UVIS2
        instrument = header['APERTURE']

    # get uncertainty file
    err_file = os.path.join(hostphot_path, 
                            'filters/HST/', 
                            f'{instrument}_err.txt')
    err_df = pd.read_csv(err_file, delim_whitespace=True)
    filt_err_df = err_df[err_df.Filter==filt]

    flux = filt_err_df.PHOTFLAM.values[0]
    flux_err = filt_err_df.ERR_PHOTFLAM.values[0]
    mag_err = np.abs(2.5 * flux_err / (flux * np.log(10)))
    
    return flux_err, mag_err

def uncertainty_calculation(
    flux, flux_err, survey, filt=None, ap_area=0.0, header=None, bkg_rms=0.0
):
    """Calculates the uncertainty propagation.

    Parameters
    ----------
    flux: float
        Aperture flux.
    flux_err: float
        Aperture flux error.
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.
    filt: str, default ``None``
        Survey-specific filter.
    ap_area: float, default ``0.0``
        Aperture area.
    header: fits header, default ``None``
        Header of an image.
    bkg_rms: float, default ``0.0``
        Background noise.

    Returns
    -------
    mag_err: float
        Extra uncertainty in magnitudes.
    flux_err: float
        Total uncertainty in flux units.
    """
    exptime = get_image_exptime(header, survey)
    gain = get_image_gain(header, survey)
    readnoise = get_image_readnoise(header, survey)

    mag_err = 2.5 / np.log(10) * flux_err / flux
    
    if survey in ["PS1", "DES", "LegacySurvey", "Spitzer", "VISTA"]:
        if survey == "Spitzer":
            flux /= header["EFCONV"]  # conv. factor (MJy/sr)/(DN/s)
        # 1.0857 = 2.5/ln(10)
        extra_err = (
            1.0857 * np.sqrt(ap_area * (readnoise**2) + flux / gain) / flux
        )
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

        extra_flux_err = np.sqrt(ap_area * (readnoise**2) + flux / gain)
        flux_err = np.sqrt(flux_err**2 + extra_flux_err**2)
    
    if survey == "DES":
        # see the photometry section in https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/quality
        # statistical uncertainties on the AB magnitud system zeropoints
        unc_dict = {
            "g": 2.6e-3,
            "r": 2.9e-3,
            "i": 3.4e-3,
            "z": 2.5e-3,
            "Y": 4.5e-3,
        }
        extra_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

        # median coadd zeropoint statistical uncertainty
        unc_dict = {
            "g": 5e-3,
            "r": 4e-3,
            "i": 5e-3,
            "z": 6e-3,
            "Y": 5e-3,
        }
        extra_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

        extra_flux_err = np.abs(flux*0.4*np.log(10)*extra_err)
        flux_err = np.sqrt(flux_err ** 2 + extra_flux_err ** 2)

    elif survey == "PS1":
        # add floor systematic error from:
        # https://iopscience.iop.org/article/10.3847/1538-4365/abb82a/pdf
        unc_dict = {
            "g": 14e-3,
            "r": 14e-3,
            "i": 15e-3,
            "z": 15e-3,
            "y": 18e-3,
        }
        floor_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err ** 2 + floor_err ** 2)

        extra_flux_err = np.abs(flux * 0.4 * np.log(10) * floor_err)
        flux_err = np.sqrt(flux_err ** 2 + extra_flux_err ** 2)

    elif survey == "SDSS":
        # https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
        camcol = header["CAMCOL"]
        run = header["RUN"]

        gain_dict = {
            "u": {1: 1.62, 2: 1.595, 3: 1.59, 4: 1.6, 5: 1.47, 6: 2.17},
            "g": {1: 3.32, 2: 3.855, 3: 3.845, 4: 3.995, 5: 4.05, 6: 4.035},
            "r": {1: 4.71, 2: 4.6, 3: 4.72, 4: 4.76, 5: 4.725, 6: 4.895},
            "i": {1: 5.165, 2: 6.565, 3: 4.86, 4: 4.885, 5: 4.64, 6: 4.76},
            "z": {1: 4.745, 2: 5.155, 3: 4.885, 4: 4.775, 5: 3.48, 6: 4.69},
        }
        # dark variance
        dv_dict = {
            "u": {
                1: 9.61,
                2: 12.6025,
                3: 8.7025,
                4: 12.6025,
                5: 9.3025,
                6: 7.0225,
            },
            "g": {
                1: 15.6025,
                2: 1.44,
                3: 1.3225,
                4: 1.96,
                5: 1.1025,
                6: 1.8225,
            },
            "r": {
                1: 1.8225,
                2: 1.00,
                3: 1.3225,
                4: 1.3225,
                5: 0.81,
                6: 0.9025,
            },
            "i": {1: 7.84, 2: 5.76, 3: 4.6225, 4: 6.25, 5: 7.84, 6: 5.0625},
            "z": {1: 0.81, 2: 1.0, 3: 1.0, 4: 9.61, 5: 1.8225, 6: 1.21},
        }
        gain = gain_dict[filt][camcol]
        dark_variance = dv_dict[filt][camcol]

        if filt == "u" and camcol == 2 and run > 1100:
            gain = 1.825
        if filt == "i" and run > 1500:
            if camcol == 2:
                dark_variance = 6.25
            if camcol == 4:
                dark_variance = 7.5625
        if filt == "z" and run > 1500:
            if camcol == 4:
                dark_variance = 12.6025
            if camcol == 5:
                dark_variance = 2.1025

        extra_err = 1.0857 * np.sqrt(dark_variance + flux / gain) / flux
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

        extra_flux_err = np.sqrt(dark_variance + flux / gain)
        flux_err = np.sqrt(flux_err ** 2 + extra_flux_err ** 2)

    elif survey == "GALEX":
        # https://asd.gsfc.nasa.gov/archive/galex/FAQ/counts_background.html
        CPS = flux
        if filt == "FUV":
            uv_err = -2.5 * (
                np.log10(CPS)
                - np.log10(
                    CPS
                    + (CPS * exptime + (0.050 * CPS * exptime) ** 2) ** 0.5
                    / exptime
                )
            )
            flux_uv_err = np.sqrt(CPS*exptime + (0.050 * CPS * exptime)**2)/exptime
        elif filt == "NUV":
            uv_err = -2.5 * (
                np.log10(CPS)
                - np.log10(
                    CPS
                    + (CPS * exptime + (0.027 * CPS * exptime) ** 2) ** 0.5
                    / exptime
                )
            )
            flux_uv_err = np.sqrt(CPS * exptime + (0.027 * CPS * exptime) ** 2) / exptime

        mag_err = np.sqrt(mag_err**2 + uv_err**2)

        flux_err = np.sqrt(flux_err ** 2 + flux_uv_err ** 2)

    elif survey == "2MASS":
        # see: https://wise2.ipac.caltech.edu/staff/jarrett/2mass/3chan/noise/
        S = flux
        N_c = 6  # number of coadd pixels
        k_z = 1.7  # kernel smoothing factor
        n_f = ap_area  # number of frame pixels in the aperture; aprox. as aperture area
        n_c = 4 * n_f  # number of coadd pixels in the aperture
        sigma_c = bkg_rms  # coadd noise; assumed to be ~background noise

        SNR = S / np.sqrt(
            S / (gain * N_c)
            + n_c * (2 * k_z * sigma_c) ** 2
            + (n_c * 0.024 * sigma_c) ** 2
        )

        mag_err = 1.0857 / SNR

        extra_flux_err = flux/SNR
        flux_err = np.sqrt(flux_err ** 2 + extra_flux_err ** 2)

    elif "WISE" in survey:
        # see: https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec2_3f.html
        # see Table 5 of
        # https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec4_4c.html#wpro
        # correction assumed to be 0 mags as PSF fitting is not used.
        m_apcor = 0.0
        f_apcor = 10 ** (-0.4 * m_apcor)
        F_src = f_apcor * flux

        N_p = (
            ap_area  # effective number of noise-pixels characterizing the PRF
        )
        # ratio of input (detector) pixel scale to output (Atlas Image) pixel scale
        pixel_scale_ratios = {"W1": 2, "W2": 2, "W3": 2, "W4": 4}
        Sin_Sout = pixel_scale_ratios[filt]
        F_corr = N_p * Sin_Sout**2

        k = 1
        N_A = N_B = ap_area
        sigma_conf = flux_err  # assumed to be the ~error in the aperture sum
        sigma_src = np.sqrt(
            f_apcor**2
            * F_corr
            * (flux_err**2 + k * (N_A**2) / N_B * bkg_rms**2)
            + sigma_conf**2
        )

        mag_err = np.sqrt(1.179 * sigma_src**2 / F_src**2)

        flux_err = f_apcor ** 2

        # add uncertainty from the ZP
        if survey=="unWISE":
            # These values are the same for all Atlas Images of a given band...
            # see: https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec2_3f.html
            unc_dict = {'W1':0.006, 'W2':0.007, 'W3':0.012, 'W4':0.012}
            zp_unc = unc_dict[filt]
        elif survey=="WISE":
            zp_unc = header["MAGZPUNC"]

        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

        extra_flux_err = np.abs(flux * 0.4 * np.log(10) * zp_unc)
        flux_err = np.sqrt(flux_err ** 2 + extra_flux_err ** 2)

    elif survey == "LegacySurvey":
        # already added at the beginning
        pass
    elif survey == "Spitzer":
        # already added at the beginning
        pass
    elif survey == "VISTA":
        # add uncertainty from the ZP
        zp_unc = header["MAGZRR"]
        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

        extra_flux_err = np.abs(flux * 0.4 * np.log(10) * zp_unc)
        flux_err = np.sqrt(flux_err ** 2 + extra_flux_err ** 2)
    elif survey == "HST":
        flux_zp_unc, zp_unc = get_HST_err(filt, header)

        mag_err = np.sqrt(mag_err**2 + zp_unc**2)
        flux_err = np.sqrt(flux_err ** 2 + flux_zp_unc ** 2)
    else:
        raise Exception(
            f"Survey {survey} has not been added for error propagation."
        )

    return mag_err, flux_err


def survey_pixel_scale(survey, filt=None):
    """Returns the pixel scale for a given survey.

    Parameters
    ----------
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.
    filt: str, default ``None``
        Filter to use by surveys that have different pixel scales
        for different filters (e.g. Spitzer's IRAC and MIPS instruments).

    Returns
    -------
    pixel_scale: float
        Pixel scale in units of arcsec/pixel.
    """
    check_survey_validity(survey)

    global config_df
    survey_df = config_df[config_df.survey == survey]
    pixel_scale = survey_df.pixel_scale.values[0]

    if len(pixel_scale.split(",")) > 1:
        filters = get_survey_filters(survey)
        pixel_scale_dict = {
            f: float(ps) for f, ps in zip(filters, pixel_scale.split(","))
        }

        if filt in pixel_scale_dict.keys():
            pixel_scale = pixel_scale_dict[filt]
        else:
            print(f"No pixel scale found for filter {filt}.")
            filt_used = list(pixel_scale_dict.keys())[0]
            print(f"Using the pixel scale of filter {filt_used} ({survey}).")
            pixel_scale = list(pixel_scale_dict.values())[0]

        return pixel_scale

    return float(pixel_scale)


def check_filters_validity(filters, survey):
    """Check whether the given filters are whithin the valid
    options for the given survey.

    Parameters
    ----------
    filters: str or list
        Filters to use, e,g, ``griz``.
    survey: str
        Survey name: e.g. ``PS1``, ``GALEX``.
    """
    if survey=='HST':
        check_HST_filters(filters)

    else:
        valid_filters = get_survey_filters(survey)

        for filt in filters:
            message = (
                f"filter '{filt}' is not a valid option for "
                f"'{survey}' survey ({valid_filters})"
            )
            assert filt in valid_filters, message

def check_HST_filters(filt):
    """Check whether the given filter is whithin the valid
    options for HST.

    Parameters
    ----------
    filt: str 
        Filter to use, e,g, ``WFC3_UVIS_F225W``.
    """        
    if filt is None:
        raise ValueError(f"'{filt}' is not a valid HST filter.")
    
    # For UVIS, only the filters of UVIS1 are used as the
    # detector 2 is scaled to match detector 1
    if 'UVIS' in filt:
        filt = filt.replace('UVIS', 'UVIS1')

    hostphot_path = hostphot.__path__[0]
    hst_file = glob.glob(os.path.join(hostphot_path, "filters/HST/*/*"))
    hst_filters = [os.path.basename(file).split('.')[0] for file in hst_file]

    assert filt in hst_filters, f"Not a valid HST filter ({filt}): {hst_filters}"


def extract_filter(filt, survey, version=None):
    """Extracts the transmission function for the filter.

    Parameters
    ----------
    filt: str
        Filter to extract.
    survey: str
        Survey of the filters.
    version: str, default ``None``
        Version of the filters to use. E.g. for the
        Legacy Survey as it uses DECam for the south
        and BASS+MzLS for the north.

    Returns
    -------
    wave: array
        Wavelength range of the filter.
    transmission: array
        Transmission function.
    """
    check_survey_validity(survey)
    if survey=='HST':
        check_filters_validity(filt, survey)
    else:
        check_filters_validity([filt], survey)

    if "WISE" in survey:
        survey = "WISE"  # for unWISE to use the same filters as WISE

    filters_path = os.path.join(hostphot.__path__[0], "filters", survey)

    # Assume DECaLS filters below 32 degrees and BASS+MzLS above
    # https://www.legacysurvey.org/status/
    if survey == "LegacySurvey":
        if version == "BASS+MzLS":
            if filt == "z":
                filt_file = os.path.join(filters_path, f"MzLS_z.dat")
            else:
                filt_file = os.path.join(filters_path, f"BASS_{filt}.dat")       
        elif version == "DECam":
            filt_file = os.path.join(filters_path, f"DECAM_{filt}.dat")

    elif survey == "HST":
        if 'UVIS' in filt:
            # Usually UVIS2 is used, but there is no large difference
            filt = filt.replace('UVIS', 'UVIS2')
        hst_files = glob.glob(os.path.join(filters_path, '*/*'))
        filt_file = [file for file in hst_files if filt in file][0]
    else:
        filt_file = os.path.join(filters_path, f"{survey}_{filt}.dat")
    
    wave, transmission = np.loadtxt(filt_file).T

    return wave, transmission


def integrate_filter(
    spectrum_wave,
    spectrum_flux,
    filter_wave,
    filter_response,
    response_type="photon",
):
    """Calcultes the flux density of an SED given a filter response.

    Parameters
    ----------
    spectrum_wave : array
        Spectrum's wavelength range.
    spectrum_flux : array
        Spectrum's flux density distribution.
    filter_wave : array
        Filter's wavelength range.
    filter_response : array
        Filter's response function.
    response_type : str, default ``photon``
        Filter's response type. Either ``photon`` or ``energy``.

    Returns
    -------
    flux_filter : float
        Flux density.
    """
    if response_type == "energy":
        filter_response = filter_response.copy() / filter_wave

    interp_response = np.interp(
        spectrum_wave, filter_wave, filter_response, left=0.0, right=0.0
    )
    int1 = np.trapz(
        spectrum_flux * interp_response * spectrum_wave, spectrum_wave
    )
    int2 = np.trapz(filter_response * filter_wave, filter_wave)
    flux_filter = int1 / int2

    return flux_filter

def adapt_aperture(objects, img_wcs, img_wcs2, flip=False):
    """Changes the aperture parameters to consider differences
    in WCS between surveys.

    The values of ``center``, ``a`` and ``b`` should in pixel
    units. DES images are flipped, so these need to be corrected
    with ``theta -> -theta``, i.e. using ``flip=True``.

    Parameters
    ----------
    objects: ndarray
        Objects with apertures.
    img_wcs: ~wcs.WCS
        WCS of the image from where the objects were extracted.
    img_wcs2: ~wcs.WCS
        WCS used to adapt the apertures.
    flip: bool, default ``False``
        Whether to flip the orientation of the aperture. Only 
        used for DES images.

    Returns
    -------
    objects_: ndarray
        Objects with adapted apertures.
    """
    objects_ = objects.copy()  # avoid modifying the intial objects
    for obj in objects_:
        center = (obj['x'], obj['y'])
        apertures = EllipticalAperture(center, obj['a'],
                                       obj['b'], obj['theta'])
        sky_apertures = apertures.to_sky(img_wcs)

        new_apertures = sky_apertures.to_pixel(img_wcs2)
        new_center = new_apertures.positions
        obj['x'], obj['y'] = new_center
        obj['a'] = new_apertures.a
        obj['b'] = new_apertures.b
        obj['theta'] = new_apertures.theta

        if flip is True:
            # flip aperture orientation
            obj['theta'] *= -1

    return objects_

def check_work_dir(wokrdir):
    """Checks if the working directory exists. If it
    does not, one is created.

    Parameters
    ----------
    wokrdir: str
        Working directory path.
    """
    if not os.path.isdir(wokrdir):
        os.mkdir(wokrdir)


def clean_dir(directory):
    """Removes the directory if it is empty.

    Parameters
    ----------
    directory: str
        Directory path.
    """
    try:
        os.rmdir(directory)
    except:
        pass


def plot_fits(fits_file, ext=0):
    """Plots a fits file.

    Parameters
    ----------
    fits_file: str or HDU.
        Path to fits file.
    ext: int
        Extension index.
    """
    if isinstance(fits_file, str):
        hdu = fits.open(fits_file)
        title = os.path.splitext(os.path.basename(fits_file))[0]
    else:
        hdu = fits_file
        title = None
    header = hdu[ext].header
    data = hdu[ext].data

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    m, s = np.nanmean(data), np.nanstd(data)

    fig = plt.figure(figsize=(10, 10))
    ax = plt.subplot(projection=img_wcs)
    ax.set_title(title, fontsize=18)
    update_axislabels(ax)

    im = ax.imshow(
        data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )
    
    plt.show()


def update_axislabels(ax):
    """Updates the labels and ticks of a plot.

    Parameters
    ----------
    ax: `.axes.SubplotBase`.
        The axis of a subplot.
    """
    for i in range(2):
        ax.coords[i].set_ticklabel(size=16)
        if i == 0:
            ax.coords[i].set_axislabel("RA (J2000)", fontsize=20)
        else:
            ax.coords[i].set_axislabel("Dec (J2000)", fontsize=20)
            # ax.coords[i].set_ticklabel(rotation=65)



import os
import glob
import copy
import shutil
import zipfile
import tarfile
import requests  # for several surveys
import numpy as np
import pandas as pd

# for VISTA
import re
import urllib

from astropy.io import fits
from astropy.table import Table
from astropy import wcs, units as u
from astropy.coordinates import SkyCoord
from astropy.nddata import Cutout2D

import pyvo  # 2MASS
from pyvo.dal import sia  # DES, DELVE
from astroquery.sdss import SDSS
from astroquery.skyview import SkyView  # other surveys
from astroquery.mast import Observations  # for GALEX

from astroquery.esa.hubble import ESAHubble
esahubble = ESAHubble()

from reproject import reproject_interp

from hostphot._constants import workdir
from hostphot.utils import (
    get_survey_filters,
    clean_dir,
    check_work_dir,
    check_survey_validity,
    check_filters_validity,
    check_HST_filters,
    survey_pixel_scale,
)

import warnings
from astropy.utils.exceptions import AstropyWarning


# PS1
# ----------------------------------------
def query_ps1(ra, dec, size=3, filters=None):
    """Query ps1filenames.py service to get a list of images

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    table: astropy Table
        Astropy table with the results.
    """
    survey = "PS1"
    check_filters_validity(filters, survey)
    if filters is None:
        filters = get_survey_filters(survey)

    pixel_scale = survey_pixel_scale(survey)
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    service = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    url = (
        f"{service}?ra={ra}&dec={dec}&size={size_pixels}&format=fits&"
        f"filters={filters}"
    )

    table = Table.read(url, format="ascii")

    return table


def get_PS1_urls(ra, dec, size=3, filters=None):
    """Get URLs for images obtained with :func:`query_ps1()`.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    url_list: list
        List of URLs for the fits images.
    """
    survey = "PS1"
    check_filters_validity(filters, survey)
    if filters is None:
        filters = get_survey_filters(survey)

    pixel_scale = survey_pixel_scale(survey)
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value

    size_pixels = int(size_arcsec / pixel_scale)

    table = query_ps1(ra, dec, size=size, filters=filters)
    url = (
        "https://ps1images.stsci.edu/cgi-bin/fitscut.cgi?"
        f"ra={ra}&dec={dec}&size={size_pixels}&format=fits"
    )

    # sort filters from blue to red
    flist = ["grizy".find(x) for x in table["filter"]]
    table = table[np.argsort(flist)]

    base_url = url + "&red="
    url_list = []
    for filename in table["filename"]:
        url_list.append(base_url + filename)

    return url_list


def get_PS1_images(ra, dec, size=3, filters=None):
    """Gets PS1 fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    fits_files: list
        List of fits images.
    """
    survey = "PS1"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    fits_url = get_PS1_urls(ra, dec, size, filters)

    hdu_list = []
    for url, filt in zip(fits_url, filters):
        hdu = fits.open(url)
        hdu_list.append(hdu)

    return hdu_list


# DES
# ----------------------------------------
def get_DES_urls(ra, dec, fov, filters="grizY"):
    """Obtains the URLs of the DES images+weights with the
    largest exposure times in the given filters.

    Parameters
    ----------
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    fov: float
        Field of view in degrees.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizY``.

    Returns
    -------
    url_list: list
        List of URLs with DES images.
    url_w_list: list
        List of URLs with DES images weights.
    """
    if filters is None:
        filters = get_survey_filters("DES")
    check_filters_validity(filters, "DES")

    des_access_url = "https://datalab.noirlab.edu/sia/des_dr1"
    svc = sia.SIAService(des_access_url)
    imgs_table = svc.search(
        (ra, dec), (fov / np.cos(dec * np.pi / 180), fov), verbosity=2
    )
    if len(imgs_table) == 0:
        print(("Warning: empty table returned for " f"ra={ra}, dec={dec}"))
        return None, None

    imgs_df = pd.DataFrame(imgs_table)

    cond_stack = imgs_df["proctype"] == "Stack"
    cond_image = imgs_df["prodtype"] == "image"
    cond_weight = imgs_df["prodtype"] == "weight"

    url_list = []
    url_w_list = []
    for filt in filters:
        cond_filt = imgs_df["obs_bandpass"] == filt

        selected_images = imgs_df[cond_filt & (cond_stack & cond_image)]
        selected_weights = imgs_df[cond_filt & (cond_stack & cond_weight)]

        if len(selected_images) > 0:
            # pick image with the longest exposure time
            max_exptime = np.argmax(selected_images["exptime"])
            row = selected_images.iloc[max_exptime]
            url = row["access_url"]  # get the download URL

            if len(selected_weights) == 0:
                url_w = None
            else:
                max_exptime = np.argmax(selected_weights["exptime"])
                row_w = selected_weights.iloc[max_exptime]
                url_w = row_w["access_url"]
        else:
            url = url_w = None
        url_list.append(url)
        url_w_list.append(url_w)

    return url_list, url_w_list


def get_DES_images(ra, dec, size=3, filters=None):
    """Gets DES fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizY``.

    Returns
    -------
    hdu_list: list
        List of fits images.
    """
    survey = "DES"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        fov = (size * u.arcmin).to(u.degree).value
    else:
        fov = size.to(u.degree).value

    url_list, url_w_list = get_DES_urls(ra, dec, fov, filters)

    if url_list is None:
        return None

    hdu_list = []
    for url, url_w in zip(url_list, url_w_list):
        # combine image+weights on a single fits file
        image_hdu = fits.open(url)
        weight_hdu = fits.open(url_w)
        hdu = fits.PrimaryHDU(image_hdu[0].data, header=image_hdu[0].header)
        hdu_err = fits.ImageHDU(
            weight_hdu[0].data, header=weight_hdu[0].header
        )
        hdu_sublist = fits.HDUList([hdu, hdu_err])
        hdu_list.append(hdu_sublist)

    return hdu_list


# SDSS
# ----------------------------------------
def get_SDSS_images(ra, dec, size=3, filters=None, version=None):
    """Downloads a set of SDSS fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``ugriz``.
    version: str, default ``None``
        Data release version to use. If not given, use the latest one (``dr17``).

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "SDSS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)
    # check data release version
    if version is None:
        version = "dr17"

    versions = [f"dr{i}" for i in range(12, 17 + 1)]
    assert (
        version in versions
    ), f"The given version ({version}) is not a valid data release: {versions}"
    dr = int(version.replace("dr", ""))

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    size_pixels = int(size_arcsec.value / pixel_scale)

    coords = SkyCoord(ra=ra * u.degree, dec=dec * u.degree)
    for radius in np.arange(1, 60, 1):
        ids = SDSS.query_region(coords, radius=radius * u.arcsec, data_release=dr)
        if ids is not None:
            break

    if ids is None:
        return None

    # get the pointing closest to the given coordinates
    coords_imgs = SkyCoord(ra=ids["ra"].value, dec=ids["dec"].value, unit=(u.degree, u.degree), frame="icrs")
    separation = coords.separation(coords_imgs).value

    pointing_id = np.argmin(separation)
    ids2remove = list(np.arange(len(separation)))
    del ids2remove[pointing_id]
    ids.remove_rows(ids2remove)

    # download images here
    hdu_list = SDSS.get_images(matches=ids, band=filters, data_release=dr)

    # SDSS images are large so need to be trimmed
    for hdu in hdu_list:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            img_wcs = wcs.WCS(hdu[0].header)

        trimmed_data = Cutout2D(hdu[0].data, coords, size_pixels, img_wcs)
        hdu[0].data = trimmed_data.data
        hdu[0].header.update(trimmed_data.wcs.to_header())

    return hdu_list


# GALEX
# ----------------------------------------
def get_GALEX_images(ra, dec, size=3, filters=None, version=None):
    """Downloads a set of GALEX fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``FUV, NUV``.
    version: str, default ``None``
        Version of GALEX images. Either Deep (``DIS``), Medium (``MIS``) or
        All-Sky Imaging Survey (``AIS``), or Nearby Galaxy Survey (``NGS``) or
        Guest Investigator Survey (``GII``). If ``None``, take the image with the
        longest exposure time.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        ``None`` is returned if no image is found.
    """
    survey = "GALEX"
    if filters is None:
        filters = get_survey_filters(survey)
    if isinstance(filters, str):
        filters = [filters]
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    size_pixels = int(size_arcsec.value / pixel_scale)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

    obs_table = Observations.query_criteria(
        coordinates=coords, radius=size_arcsec, obs_collection=["GALEX"]
    )
    obs_df_ = obs_table.to_pandas()

    if version is None:
        # starts from the survey with the deepest images first
        obs_df_ = obs_df_.sort_values("t_exptime", ascending=False)
        projects = obs_df_.project.unique()
    else:
        # only use the survey requested by the user
        projects = [version]

    hdu_dict = {"NUV": None, "FUV": None}
    for project in projects:
        obs_df = obs_df_[obs_df_.project == project]

        for filt in filters:
            if hdu_dict[filt] is not None:
                # if the image was already found, skip this filter
                continue

            # get only "intensity" images
            filt_extension = {
                "NUV": "-nd-int.fits.gz",
                "FUV": "-fd-int.fits.gz",
            }
            # get unique image sectors
            files = []
            for file in obs_df.dataURL.values:
                sector_info = file.split("-")[:-2]
                file = (
                    "-".join(string for string in sector_info)
                    + filt_extension[filt]
                )
                if file not in files:
                    files.append(file)

            # download the FITS images
            hdu_list = []
            for file in files:
                try:
                    hdu = fits.open(file)
                    hdu_list.append(hdu)
                except:
                    pass

            # calculate the separation of the galaxy to the image center
            separations = []
            for hdu in hdu_list:
                ra_img = float(hdu[0].header["RA_CENT"])
                dec_img = float(hdu[0].header["DEC_CENT"])
                coords_img = SkyCoord(ra=ra_img, dec=dec_img, unit=(u.degree, u.degree), frame="icrs")
                separation = coords.separation(coords_img).value
                separations.append(separation)

            # get the image with the galaxy closest to the center
            if len(separations) != 0:
                id_file = np.argmin(separations)
                hdu = hdu_list[id_file]
            else:
                hdu = None

            hdu_dict[filt] = hdu

    hdu_list = []
    for filt in filters:
        hdu = hdu_dict[filt]
        if hdu is None:
            continue  # no image in this filter
        # trim data to requested size
        img_wcs = wcs.WCS(hdu[0].header)
        pos = SkyCoord(ra=ra * u.degree, dec=dec * u.degree)

        trimmed_data = Cutout2D(hdu[0].data, pos, size_pixels, img_wcs)
        hdu[0].data = trimmed_data.data
        hdu[0].header.update(trimmed_data.wcs.to_header())
        hdu_list.append(hdu)

    return hdu_list


# WISE
# ----------------------------------------
def get_used_image(header):
    """Obtains the name of the image downloaded by SkyView.

    Parameters
    ----------
    header: fits header
        Header of an image.

    Returns
    -------
    used_image: str
        Name of the image.
    """
    image_line = header["HISTORY"][-3]
    used_image = image_line.split("/")[-1].split("-")[0]

    return used_image


def get_WISE_images(ra, dec, size=3, filters=None):
    """Downloads a set of WISE fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses all WISE filters.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "WISE"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    # size_pixels = int(size_arcsec.value / pixel_scale)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
        # just to get the original image used
        skyview_fits = SkyView.get_images(
            position=coords,
            coordinates="icrs",
            pixels="100",
            survey="WISE 3.4",
        )
        header = skyview_fits[0][0].header

        coadd_id = get_used_image(header)
        coadd_id1 = coadd_id[:4]
        coadd_id2 = coadd_id1[:2]

        # for more info: https://irsa.ipac.caltech.edu/ibe/docs/wise/allwise/p3am_cdd/#sample_code
        base_url = (
            "http://irsa.ipac.caltech.edu/ibe/data/wise/allwise/p3am_cdd/"
        )
        coadd_url = os.path.join(coadd_id2, coadd_id1, coadd_id)
        params_url = f"center={ra},{dec}&size={size_arcsec.value}arcsec&gzip=0"  # center and size of the image

        hdu_list = []
        for filt in filters:
            i = filt[-1]
            band_url = f"{coadd_id}-w{i}-int-3.fits"
            url = os.path.join(
                base_url, coadd_url, band_url + "?" + params_url
            )
            hdu = fits.open(url)
            hdu_list.append(hdu)

    return hdu_list


def get_unWISE_images(ra, dec, size=3, filters=None, version="allwise"):
    """Downloads a set of unWISE fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses all WISE filters.
    version: str, default ``allwise``
        Version of the unWISE images. Either ``allwise`` or ``neo{i}`` for
        {i} = 1 to 7.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "unWISE"
    if version is None:
        version = "allwise"
    else:
        # check validity of the version used
        neo_versions = [f"neo{i}" for i in range(1, 8)]
        all_versions = ["allwise"] + neo_versions
        assert (
            version in all_versions
        ), f"Not a valid version ({version}): {all_versions}"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if "neo" in version:
        # these only have W1 and W2 data
        if "W3" in filters:
            filters.remove("W3")
        if "W4" in filters:
            filters.remove("W4")

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    size_pixels = int(size_arcsec.value / pixel_scale)
    assert size_pixels <= 1024, "Maximum cutout size for unWISE is 1024 pixels"

    bands = "".join(filt[-1] for filt in filters)  # e.g. 1234

    # for more info: http://unwise.me/imgsearch/
    base_url = "http://unwise.me/cutout_fits?"
    params_url = (
        f"version={version}&ra={ra}&dec={dec}&size={size_pixels}&bands={bands}"
    )
    master_url = base_url + params_url

    response = requests.get(master_url, stream=True)
    target_file = f"unWISE_images_{ra}_{dec}.tar.gz"  # current directory
    if response.status_code == 200:
        with open(target_file, "wb") as f:
            f.write(response.raw.read())

    hdu_list = []
    with tarfile.open(target_file) as tar_file:
        files_list = tar_file.getnames()
        for fits_file in files_list:
            for filt in filters:
                if f"{filt.lower()}-img-m.fits" in fits_file:
                    tar_file.extract(fits_file, ".")
                    hdu = fits.open(fits_file)
                    hdu_list.append(hdu)
                    os.remove(fits_file)

    # sometimes, the file is not downloaded, so let's check if it exists
    if os.path.isfile(target_file) is True:
        os.remove(target_file)

    return hdu_list


# 2MASS
# ----------------------------------------
def get_2MASS_images(ra, dec, size=3, filters=None):
    """Downloads a set of 2MASS fits images for a given set
    of coordinates and filters.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``FUV, NUV``.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "2MASS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_degree = (size * u.arcmin).to(u.degree)
    else:
        size_degree = size.to(u.degree)
    size_degree = size_degree.value

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

        twomass_services = pyvo.regsearch(
            servicetype="image", keywords=["2mass"]
        )
        table = twomass_services[0].search(pos=coords, size=size_degree)
        twomass_df = table.to_table().to_pandas()
        twomass_df = twomass_df[twomass_df.format == "image/fits"]

        # for more info: https://irsa.ipac.caltech.edu/ibe/docs/twomass/allsky/allsky/#main
        base_url = (
            "https://irsa.ipac.caltech.edu/ibe/data/twomass/allsky/allsky"
        )

        hdu_list = []
        for filt in filters:
            if filt == "Ks":
                filt = "K"
            band_df = twomass_df[twomass_df.band == filt]
            if len(band_df) == 0:
                # no data for this band:
                hdu_list.append(None)
                continue

            # get image with the image's coordinates closest to the given coordinates
            coords_imgs = SkyCoord(ra=band_df.center_ra.values, 
                                   dec=band_df.center_dec.values, unit=(u.degree, u.degree), frame="icrs")
            separation = coords.separation(coords_imgs).value
            i = np.argmin(separation)

            fname = band_df.download.values[i].split("=")[-1]
            hemisphere = band_df.hem.values[i]
            ordate = band_df.date.values[i]
            scanno = band_df.scan.values[i]
            # add leading zeros for scanno bellow 100
            n_zeros = 3 - len(str(scanno))
            scanno = n_zeros * "0" + str(scanno)

            tile_url = os.path.join(f"{ordate}{hemisphere}", f"s{scanno}")
            fits_url = os.path.join("image", f"{fname}.gz")
            params_url = f"center={ra},{dec}&size={size_degree}degree&gzip=0"  # center and size of the image

            url = os.path.join(base_url, tile_url, fits_url + "?" + params_url)
            hdu = fits.open(url)
            hdu_list.append(hdu)

    return hdu_list

# Legacy Survey
def get_LegacySurvey_images(ra, dec, size=3, filters=None, version="dr10"):
    """Gets Legacy Survey fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grz``.
    version: str, default ``dr10``
        Data release version. E.g. ``dr10``, ``dr9``, ``dr8``, etc..

    Returns
    -------
    fits_files: list
        List of fits images.
    """
    survey = "LegacySurvey"
    if filters is None:
        filters = get_survey_filters(survey)
        check_filters_validity(filters, survey)

    pixel_scale = survey_pixel_scale(survey)
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    if version is None:
        version = "dr10"  # latest data release

    base_url = "https://www.legacysurvey.org/viewer/fits-cutout?"
    params = f"ra={ra}&dec={dec}&layer=ls-{version}&pixscale={pixel_scale}&bands={filters}&size={size_pixels}"
    url = base_url + params
    master_hdu = fits.open(url)
    header = master_hdu[0].header

    hdu_list = []
    # assuming order grz of the filters
    for i, filt in enumerate(filters):
        data = master_hdu[0].data[i]
        hdu = fits.PrimaryHDU(data=data, header=header)
        hdu_list.append(hdu)

    return hdu_list
    

# Spitzer
def get_Spitzer_images(ra, dec, size=3, filters=None):
    """Gets Spitzer fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``IRAC.1, IRAC.2, IRAC.3, IRAC.4, MIPS.1``.

    Returns
    -------
    fits_files: list
        List of fits images.
    """
    survey = "Spitzer"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    pixel_scale = survey_pixel_scale(survey, "IRAC.1")  # IRAC resolution
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    base_url = "https://irsa.ipac.caltech.edu/cgi-bin/Cutouts/nph-cutouts?"
    locstr = f"{str(ra)}+{str(dec)}+eq"

    mission_dict = {
        "mission": "SEIP",
        "min_size": "18",
        "max_size": "1800",
        "units": "arcsec",
        "locstr": locstr,
        "sizeX": size_pixels,
        "ntable_cutouts": "1",
        "cutouttbl1": "science",
        "mode": "PI",  # XML output file
    }

    # check image size
    if size_pixels < int(mission_dict["min_size"]):
        mission_dict["sizeX"] = int(mission_dict["min_size"])
        print(
            f'Image cannot be smaller than {mission_dict["min_size"]} arcsec (using this value)'
        )
    elif size_pixels > int(mission_dict["max_size"]):
        mission_dict["sizeX"] = int(mission_dict["max_size"])
        print(
            f'Image cannot be larger than {mission_dict["max_size"]} arcsec (using this value)'
        )

    mission_url = "&".join(
        [f"{key}={val}" for key, val in mission_dict.items()]
    )
    url = base_url + mission_url

    response = requests.get(url)
    list_text = response.text.split()

    # find the fits file for each fitler (None if not found)
    files_dict = {filt: None for filt in filters}
    for filt in filters:
        for line in list_text:
            if filt in line and line.endswith(".mosaic.fits"):
                files_dict[filt] = line
                # pick the first image and move to the next filter
                break

    hdu_list = []
    for filt, file in files_dict.items():
        if file is not None:
            hdu = fits.open(line)
            hdu[0].header["MAGZP"] = hdu[0].header["ZPAB"]
            hdu_list.append(hdu)
        else:
            hdu_list.append(None)

    return hdu_list


# VISTA
def get_VISTA_images(ra, dec, size=3, filters=None, version="VHS"):
    """Gets VISTA fits images for the given coordinates and
    filters.

    Note: the different surveys included in VISTA cover different
    parts of the sky and do not necessarily contain the same filters.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``Z, Y, J, H, Ks``.
    version: str, default ``VHS``
        Survey to use: ``VHS``, ``VIDEO`` or ``VIKING``.

    Returns
    -------
    fits_files: list
        List of fits images.
    """
    survey = "VISTA"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if not isinstance(size, (float, int)):
        size = size.to(u.arcmin).value

    if version is None:
        version = "VHS"
    # These are final data releases except for VHS(?):
    # VHSDR5: https://www.eso.org/sci/publications/announcements/sciann17290.html
    # VHSDR6: https://b2find.eudat.eu/dataset/0b10d3a0-1cfe-5e67-8a5c-0949db9d19cb
    # VIDEODR5: https://www.eso.org/sci/publications/announcements/sciann17491.html
    # VIKINGDR4: https://www.eso.org/sci/publications/announcements/sciann17289.html
    database_dict = {
        "VHS": "VHSDR6",
        "VIDEO": "VIDEODR5",
        "VIKING": "VIKINGDR4",
    }
    valid_surveys = list(database_dict.keys())
    assert (
        version in valid_surveys
    ), f"Not a valid VISTA survey: choose from {valid_surveys}"
    database = database_dict[version]

    base_url = "http://horus.roe.ac.uk:8080/vdfs/GetImage?archive=VSA&"
    survey_dict = {
        "database": database,
        "ra": ra,
        "dec": dec,
        "sys": "J",
        "filterID": "all",
        "size": size,  # in arcmin
        "obsType": "object",
        "frameType": "tilestack",
    }
    survey_url = "&".join([f"{key}={val}" for key, val in survey_dict.items()])

    url = base_url + survey_url
    results = urllib.request.urlopen(url).read()
    links = re.findall('href="(http://.*?)"', results.decode("utf-8"))

    # find url for each filter (None if not found)
    urls_dict = {filt: None for filt in filters}
    for filt in filters:
        for link in links:
            url = link.replace("getImage", "getFImage", 1)
            if f"band={filt}" in url:
                urls_dict[filt] = url
                break

    hdu_list = []
    for filt, url in urls_dict.items():
        if url is not None:
            hdu = fits.open(url)
            # to be consistent with the naming of other surveys
            hdu[1].header["MAGZP"] = hdu[1].header["MAGZPT"]
            hdu_list.append(hdu[1])  # extension 1
        else:
            hdu_list.append(None)

    return hdu_list


# HST
# ----------------------------------------
def update_HST_header(hdu):
    """Updates the HST image header with the necessary keywords.

    Parameters
    ----------
    hdu : Header Data Unit.
        HST FITS image.
    """
    # get WCS
    with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            for i in range(1, len(hdu)-1):
                try:
                    img_wcs = wcs.WCS(hdu[i].header)
                except:
                    continue
    hdu[0].header.update(img_wcs.to_header())
    hdu[0].header['PHOTFLAM'] = hdu[1].header['PHOTFLAM']
    hdu[0].header['PHOTPLAM'] = hdu[1].header['PHOTPLAM']
    hdu[0].data = hdu[1].data

    # add zeropoints
    # https://www.stsci.edu/hst/instrumentation/acs/data-analysis/zeropoints
    photflam = hdu[0].header["PHOTFLAM"]
    photplam = hdu[0].header["PHOTPLAM"]
    hdu[0].header["MAGZP"] = (
        -2.5 * np.log10(photflam) - 5 * np.log10(photplam) - 2.408
    )

def set_HST_image(file, filt, name):
    """Moves a previously downloaded HST image into the work directory.

    The image's header is updated with the necessary keywords to obtain
    photometry and is also moved under the objects directory inside the 
    work directory.

    HST images take very long to download, so the user might prefer to
    download the images manually and then use this function to include
    the image into the workflow.

    Parameters
    ----------
    file : str
        HST image to use.
    filt : str
        HST filter, e.g. ``WFC3_UVIS_F275W``.
    name : str
        Object's name.
    """
    check_work_dir(workdir)
    obj_dir = os.path.join(workdir, name)
    if not os.path.isdir(obj_dir):
        os.mkdir(obj_dir)

    hdu = fits.open(file)
    update_HST_header(hdu)
    outfile = os.path.join(obj_dir, f'HST_{filt}.fits')
    hdu.writeto(outfile, overwrite=True)

def get_HST_images(ra, dec, size=3, filt=None):
    """Downloads a set of HST fits images for a given set
    of coordinates and filters using the MAST archive.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filt: str, default ``None``
        Filter to use, e.g. ``WFC3_UVIS_F225W``.

    Return
    ------
    hdu_list: list
        List with fits image for the given filter.
        `None` is returned if no image is found.
    """
    global esahubble
    esahubble.get_status_messages()
    check_HST_filters(filt)

    # separate the instrument name from the actual filter
    split_filt = filt.split('_')
    if len(split_filt)==2:
        filt = split_filt[-1]
        instrument = split_filt[0]
    elif len(split_filt)==3:
        filt = split_filt[-1]
        instrument = f'{split_filt[0]}/{split_filt[1]}'
    else:
        raise ValueError(f"Incorrect filter name: {filt}")

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)
    size_arcsec = size_arcsec.value

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

    version = None
    if version=='HLA':
        # This does not seem to be faster
        fov = 0.2   # field-of-view win degrees
        access_url = " https://hla.stsci.edu/cgi-bin/hlaSIAP.cgi"
        svc = sia.SIAService(access_url)
        imgs_table = svc.search(
            (ra, dec), (fov / np.cos(dec * np.pi / 180), fov), verbosity=2
        )

        obs_df = pd.DataFrame(imgs_table)
        obs_df = obs_df[obs_df.Mode=='IMAGE']
        obs_df = obs_df[obs_df.Format.str.endswith('fits')]
        obs_df = obs_df[obs_df.Detector==instrument]
        obs_df = obs_df[obs_df.Spectral_Elt==filt]
        obs_df = obs_df[obs_df.ExpTime==obs_df.ExpTime.max()]
        
        hdu = fits.open(obs_df.URL.values[0])
    else:
        result = esahubble.cone_search_criteria(radius=3,
                                                coordinates=coords,
                                                calibration_level='PRODUCT',
                                                data_product_type = 'image',
                                                instrument_name = instrument,
                                                filters = filt,
                                                async_job = True,
                                            )
        
        obs_df = result.to_pandas()
        obs_df = obs_df[obs_df['filter']==filt]
        # get only exposures shorter than one hour
        obs_df = obs_df[obs_df.exposure_duration<3600]  
        obs_df.sort_values(by=['exposure_duration'], 
                        ascending=False, inplace=True) 

        print('Looking for HST images...')
        filename = f'HST_tmp_{ra}_{dec}'  # the extension is added below
        for obs_id in obs_df.observation_id:
            try:
                esahubble.download_product(observation_id=obs_id, 
                                        product_type="SCIENCE", 
                                        calibration_level="PRODUCT", 
                                        filename=filename)
                break
            except:
                pass

        temp_file = f'{filename}.fits.gz'
        if os.path.isfile(temp_file) is False:
            return None
        
        temp_dir = filename # same name as the extension-less file above
        with zipfile.ZipFile(temp_file, 'r') as zip_ref:
            zip_ref.extractall(temp_dir)
            fits_file = [file for file in glob.glob(f'{temp_dir}/**', recursive=True) 
                        if file.endswith('.gz')][0]

        hdu = fits.open(fits_file)    

        # remove the temporary files and directory
        os.remove(temp_file) 
        shutil.rmtree(temp_dir, ignore_errors=True)

    update_HST_header(hdu)
    hdu_list = [hdu]
    
    return hdu_list

def get_HST_images_OLD(ra, dec, size=3, filt=None, instrument=None):
    """Downloads a set of HST fits images for a given set
    of coordinates and filters using the MAST archive.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filt: str, default ``None``
        Filter to use.
    instrument: str, default ``None``
        Instrument to use.

    Return
    ------
    hdu_list: list
        List with fits image for the given filter.
        `None` is returned if no image is found.
    """
    check_HST_filters(filt, instrument)

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)
    size_arcsec = size_arcsec.value

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

    obs_table = Observations.query_region(coords, radius=size)
    obs_table = obs_table.filled()  # remove masked rows
    obs_df = obs_table.to_pandas()

    obs_df = obs_df[
        (obs_df.obs_collection == "HST") | (obs_df.obs_collection == "HLA")
    ]
    obs_df = obs_df[obs_df.dataproduct_type == "image"]
    obs_df = obs_df[obs_df.t_exptime > 15]  # remove short exposures

    # filter by instrument+filter
    inst = instrument.split("/")[0]
    # inst_df = obs_df[obs_df.instrument_name==instrument]
    inst_df = obs_df[obs_df.instrument_name.str.contains(inst)]
    filt_df = inst_df[inst_df.filters == filt]
    # just use one image
    filt_df = filt_df[filt_df.t_exptime == filt_df.t_exptime.max()]

    # get data products
    data_products = Observations.get_product_list(Table.from_pandas(filt_df))
    dp_df = data_products.to_pandas()
    dp_df = dp_df[dp_df.type == "S"]
    dp_df = dp_df[dp_df.productSubGroupDescription == "FLT"]  # only images
    # choose first image
    single_dp_df = dp_df[dp_df.obs_id == dp_df.obs_id.values[0]]

    Observations.download_products(
        Table.from_pandas(single_dp_df),
        download_dir=None,
        cache=False,
        productType="SCIENCE",
        extension=["fits"],
    )

    for path, subdirs, files in os.walk("mastDownload"):
        for file in files:
            fits_file = os.path.join(path, file)

    hdu = fits.open(fits_file)
    hdu[0].data = hdu[1].data

    # add zeropoints
    # https://www.stsci.edu/hst/instrumentation/acs/data-analysis/zeropoints
    photflam = hdu[0].header["PHOTFLAM"]
    photplam = hdu[0].header["PHOTFLAM"]
    hdu[0].header["MAGZP"] = (
        -2.5 * np.log10(photflam) - 5 * np.log10(photplam) - 2.408
    )
    hdu_list = [hdu]

    # remove directory created by MAST download
    shutil.rmtree("mastDownload", ignore_errors=True)

    # HST images can be large so need to be trimmed
    pixel_scale = survey_pixel_scale('HST')
    size_pixels = int(size_arcsec / pixel_scale)
    pos = SkyCoord(ra=ra * u.degree, dec=dec * u.degree)

    for hdu in hdu_list:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", AstropyWarning)
            img_wcs = wcs.WCS(hdu[0].header)

        trimmed_data = Cutout2D(hdu[0].data, pos, size_pixels, img_wcs)
        hdu[0].data = trimmed_data.data
        hdu[0].header.update(trimmed_data.wcs.to_header())

    return hdu_list


# Check orientation
# ------------------
def match_wcs(fits_files):
    """Matches the WCS of all the images, taking the first ones as
    a reference.

    Parameters
    ----------
    fits_files: list
        List of fits files.

    Returns
    -------
    matched_fits_files: list
        List of fits files with matched WCS.
    """
    matched_fits_files = copy.deepcopy(fits_files)
    # some hdu have data + error
    hdu0 = matched_fits_files[0][0]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        wcs0 = wcs.WCS(hdu0.header)

        for hdu in matched_fits_files[1:]:
            data, footprint = reproject_interp(hdu[0], hdu0.header)
            hdu[0].data = data
            hdu[0].header.update(wcs0.to_header())

    return matched_fits_files


# Master Function
# ----------------------------------------
def download_images(
    name,
    ra,
    dec,
    size=3,
    filters=None,
    overwrite=True,
    survey="PS1",
    version=None,
):
    """Download images for a given object in the given filters of a
    given survey.

    The surveys that use the ``version`` parameter are: GALEX (``AIS``, ``MIS``,
    ``DIS``, ``NGS`` and ``GII``),  unWISE (``allwise`` and ``neo{i}`` for {i}=1-7),
    VISTA (``VHS``, ``VIDEO`` and ``VIKING``) and SDSS (``dr{i}`` for {i}=12-17).

    Parameters
    ----------
    name: str
        Name used for tracking the object in your local
        directory.
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters for the images.
    overwrite: bool, default ``True``
        If ``True``, the images are overwritten if they already
        exist.
    survey: str, default ``PS1``
        Survey used to download the images.
    version: str, default ``None``
        Version used by some surveys including multiple surveys. E.g. ``VHS`` for VISTA.

    Examples
    --------
    >>> from hostphot.cutouts import download_images
    >>> name = 'SN2004eo'
    >>> host_ra, host_dec = 308.2092, 9.92755  # coords of host galaxy of SN2004eo
    >>> survey = 'PS1'
    >>> download_images(name, host_ra, host_dec, survey=survey)
    """
    check_survey_validity(survey)

    check_work_dir(workdir)
    obj_dir = os.path.join(workdir, name)
    if not os.path.isdir(obj_dir):
        os.mkdir(obj_dir)

    # download fits files
    if survey == "PS1":
        hdu_list = get_PS1_images(ra, dec, size, filters)
    elif survey == "DES":
        hdu_list = get_DES_images(ra, dec, size, filters)
    elif survey == "SDSS":
        hdu_list = get_SDSS_images(ra, dec, size, filters, version)
    elif survey == "GALEX":
        hdu_list = get_GALEX_images(ra, dec, size, filters, version)
    elif survey == "WISE":
        hdu_list = get_WISE_images(ra, dec, size, filters)
    elif survey == "unWISE":
        hdu_list = get_unWISE_images(ra, dec, size, filters, version)
    elif survey == "2MASS":
        hdu_list = get_2MASS_images(ra, dec, size, filters)
    elif survey == "HST":
        hdu_list = get_HST_images(ra, dec, size, filters)
    elif survey == "LegacySurvey":
        hdu_list = get_LegacySurvey_images(ra, dec, size, filters, version)
    elif survey == "Spitzer":
        hdu_list = get_Spitzer_images(ra, dec, size, filters)
    elif survey == "VISTA":
        hdu_list = get_VISTA_images(ra, dec, size, filters, version)
    else:
        raise ValueError(
            "The given survey is not properly added to HostPhot..."
        )
    
    if filters is None:
        filters = get_survey_filters(survey)

    if hdu_list:
        for hdu, filt in zip(hdu_list, filters):
            if hdu is None:
                continue  # skip missing filter/image

            if survey == "HST":
                inst = version.replace("/", "-")
                outfile = os.path.join(
                    obj_dir, f"{survey}_{inst}-{filters}.fits"
                )
            else:
                outfile = os.path.join(obj_dir, f"{survey}_{filt}.fits")

            if overwrite is True or os.path.isfile(outfile) is False:
                hdu.writeto(outfile, overwrite=overwrite)
            else:
                continue

    # remove directory if it ends up empty
    clean_dir(obj_dir)



import unittest
from hostphot.cutouts import download_images
from hostphot.rgb_images import (
    create_RGB_image,
    get_PS1_RGB_image,
    get_SDSS_RGB_image,
)


class TestHostPhot(unittest.TestCase):
    def test_create_RGB_image(self):
        sn_name = "2002fk"
        ra = 50.527333
        dec = -15.400056
        download_images(sn_name, ra, dec, survey="PS1")

        images_dir = "images/2002fk"
        for scaling in ["linear", "sqrt", "log", "asinh"]:
            create_RGB_image(images_dir=images_dir, scaling=scaling)

    def test_get_PS1_RGB_image(self):
        ra, dec = 50.527333, -15.400056
        get_PS1_RGB_image("PS1.jpg", ra=ra, dec=dec)

    #def test_get_SDSS_RGB_image(self):
    #    ra, dec = 50.527333, -15.400056
    #    get_SDSS_RGB_image("SDSS.jpg", ra=ra, dec=dec)


if __name__ == "__main__":
    unittest.main()
